import unittest
import libpysal
import numpy as np
from scipy import sparse
from spreg import error_sp_het as HET
from libpysal.common import RTOL
import spreg


class TestBaseGMErrorHet(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("CRIME"))
        self.X = np.array(X).T
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        self.X = sparse.csr_matrix(self.X)
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HET.BaseGM_Error_Het(self.y, self.X, self.w.sparse, step1c=True)
        betas = np.array([[47.99626638], [0.71048989], [-0.55876126], [0.41178776]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([27.38122697])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        ef = np.array([32.29765975])
        np.testing.assert_allclose(reg.e_filtered[0], ef, RTOL)
        predy = np.array([53.08577603])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n)
        k = 3
        np.testing.assert_allclose(reg.k, k)
        y = np.array([80.467003])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.x[0].toarray()[0], x, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        stdy = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, stdy)
        vm = np.array(
            [
                [1.31767529e02, -3.58368748e00, -1.65090647e00, 0.00000000e00],
                [-3.58368748e00, 1.35513711e-01, 3.77539055e-02, 0.00000000e00],
                [-1.65090647e00, 3.77539055e-02, 2.61042702e-02, 0.00000000e00],
                [0.00000000e00, 0.00000000e00, 0.00000000e00, 2.82398517e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        xtx = np.array(
            [
                [4.90000000e01, 7.04371999e02, 1.72131237e03],
                [7.04371999e02, 1.16866734e04, 2.15575320e04],
                [1.72131237e03, 2.15575320e04, 7.39058986e04],
            ]
        )
        np.testing.assert_allclose(reg.xtx, xtx, RTOL)


class TestGMErrorHet(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("CRIME"))
        self.X = np.array(X).T
        self.X = sparse.csr_matrix(self.X)
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HET.GM_Error_Het(self.y, self.X, self.w, step1c=True)
        betas = np.array([[47.99626638], [0.71048989], [-0.55876126], [0.41178776]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([27.38122697])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        ef = np.array([32.29765975])
        np.testing.assert_allclose(reg.e_filtered[0], ef, RTOL)
        predy = np.array([53.08577603])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n)
        k = 3
        np.testing.assert_allclose(reg.k, k)
        y = np.array([80.467003])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.x[0].toarray()[0], x, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        stdy = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, stdy)
        vm = np.array(
            [
                [1.31767529e02, -3.58368748e00, -1.65090647e00, 0.00000000e00],
                [-3.58368748e00, 1.35513711e-01, 3.77539055e-02, 0.00000000e00],
                [-1.65090647e00, 3.77539055e-02, 2.61042702e-02, 0.00000000e00],
                [0.00000000e00, 0.00000000e00, 0.00000000e00, 2.82398517e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        pr2 = 0.34951013222581306
        np.testing.assert_allclose(reg.pr2, pr2)
        stde = np.array([11.47900385, 0.36812187, 0.16156816, 0.16804717])
        np.testing.assert_allclose(reg.std_err, stde, RTOL)
        z_stat = np.array(
            [
                [4.18122226e00, 2.89946274e-05],
                [1.93003988e00, 5.36018970e-02],
                [-3.45836247e00, 5.43469673e-04],
                [2.45042960e00, 1.42685863e-02],
            ]
        )
        np.testing.assert_allclose(reg.z_stat, z_stat, RTOL)
        xtx = np.array(
            [
                [4.90000000e01, 7.04371999e02, 1.72131237e03],
                [7.04371999e02, 1.16866734e04, 2.15575320e04],
                [1.72131237e03, 2.15575320e04, 7.39058986e04],
            ]
        )
        np.testing.assert_allclose(reg.xtx, xtx, RTOL)


class TestBaseGMEndogErrorHet(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        self.X = sparse.csr_matrix(self.X)
        yd = []
        yd.append(db.by_col("CRIME"))
        self.yd = np.array(yd).T
        q = []
        q.append(db.by_col("DISCBD"))
        self.q = np.array(q).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HET.BaseGM_Endog_Error_Het(
            self.y, self.X, self.yd, self.q, self.w.sparse, step1c=True
        )
        betas = np.array([[55.39707924], [0.46563046], [-0.67038326], [0.41135023]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([26.51812895])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        ef = np.array([31.46604707])
        np.testing.assert_allclose(reg.e_filtered[0], ef, RTOL)
        predy = np.array([53.94887405])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n)
        k = 3
        np.testing.assert_allclose(reg.k, k)
        y = np.array([80.467003])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531])
        np.testing.assert_allclose(reg.x[0].toarray()[0], x, RTOL)
        yend = np.array([15.72598])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([5.03])
        np.testing.assert_allclose(reg.q[0], q, RTOL)
        z = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.z[0].toarray()[0], z, RTOL)
        h = np.array([1.0, 19.531, 5.03])
        np.testing.assert_allclose(reg.h[0].toarray()[0], h, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        stdy = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, stdy)
        vm = np.array(
            [
                [8.34637805e02, -2.16932259e01, -1.33327894e01, 1.65840848e00],
                [-2.16932259e01, 5.97683070e-01, 3.39503523e-01, -3.90111107e-02],
                [-1.33327894e01, 3.39503523e-01, 2.19008080e-01, -2.81929695e-02],
                [1.65840848e00, -3.90111107e-02, -2.81929695e-02, 3.15686105e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        hth = np.array(
            [
                [49.0, 704.371999, 139.75],
                [704.371999, 11686.67338121, 2246.12800625],
                [139.75, 2246.12800625, 498.5851],
            ]
        )
        np.testing.assert_allclose(reg.hth, hth, RTOL)


class TestGMEndogErrorHet(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        self.X = sparse.csr_matrix(self.X)
        yd = []
        yd.append(db.by_col("CRIME"))
        self.yd = np.array(yd).T
        q = []
        q.append(db.by_col("DISCBD"))
        self.q = np.array(q).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HET.GM_Endog_Error_Het(
            self.y, self.X, self.yd, self.q, self.w, step1c=True
        )
        betas = np.array([[55.39707924], [0.46563046], [-0.67038326], [0.41135023]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([26.51812895])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([53.94887405])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n)
        k = 3
        np.testing.assert_allclose(reg.k, k)
        y = np.array([80.467003])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531])
        np.testing.assert_allclose(reg.x[0].toarray()[0], x, RTOL)
        yend = np.array([15.72598])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([5.03])
        np.testing.assert_allclose(reg.q[0], q, RTOL)
        z = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.z[0].toarray()[0], z, RTOL)
        h = np.array([1.0, 19.531, 5.03])
        np.testing.assert_allclose(reg.h[0].toarray()[0], h, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        stdy = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, stdy)
        vm = np.array(
            [
                [8.34637805e02, -2.16932259e01, -1.33327894e01, 1.65840848e00],
                [-2.16932259e01, 5.97683070e-01, 3.39503523e-01, -3.90111107e-02],
                [-1.33327894e01, 3.39503523e-01, 2.19008080e-01, -2.81929695e-02],
                [1.65840848e00, -3.90111107e-02, -2.81929695e-02, 3.15686105e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        pr2 = 0.34648011338954804
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        std_err = np.array([28.89009873, 0.77309965, 0.46798299, 0.17767558])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        z_stat = np.array(
            [
                (1.9175109006819244, 0.055173057472126787),
                (0.60229035155742305, 0.54698088217644414),
                (-1.4324949211864271, 0.15200223057569454),
                (2.3151759776869496, 0.020603303355572443),
            ]
        )
        np.testing.assert_allclose(reg.z_stat, z_stat, RTOL)
        hth = np.array(
            [
                [49.0, 704.371999, 139.75],
                [704.371999, 11686.67338121, 2246.12800625],
                [139.75, 2246.12800625, 498.5851],
            ]
        )
        np.testing.assert_allclose(reg.hth, hth, RTOL)


class TestBaseGMComboHet(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("CRIME"))
        self.X = np.array(X).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        # Only spatial lag
        yd2, q2 = spreg.utils.set_endog(self.y, self.X, self.w, None, None, 1, True)
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        self.X = sparse.csr_matrix(self.X)
        reg = HET.BaseGM_Combo_Het(
            self.y, self.X, yend=yd2, q=q2, w=self.w.sparse, step1c=True
        )
        betas = np.array(
            [[57.7778574], [0.73034922], [-0.59257362], [-0.2230231], [0.56636724]]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([25.65156033])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        ef = np.array([31.87664403])
        np.testing.assert_allclose(reg.e_filtered[0], ef, RTOL)
        predy = np.array([54.81544267])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n)
        k = 4
        np.testing.assert_allclose(reg.k, k)
        y = np.array([80.467003])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.x[0].toarray()[0], x, RTOL)
        yend = np.array([35.4585005])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([18.594, 24.7142675])
        np.testing.assert_allclose(reg.q[0], q, RTOL)
        z = np.array([1.0, 19.531, 15.72598, 35.4585005])
        np.testing.assert_allclose(reg.z[0].toarray()[0], z, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        stdy = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, stdy, RTOL)
        vm = np.array(
            [
                [
                    4.86218274e02,
                    -2.77268729e00,
                    -1.59987770e00,
                    -1.01969471e01,
                    2.74302006e00,
                ],
                [
                    -2.77268729e00,
                    1.04680972e-01,
                    2.51172238e-02,
                    1.95136385e-03,
                    3.70052723e-03,
                ],
                [
                    -1.59987770e00,
                    2.51172238e-02,
                    2.15655720e-02,
                    7.65868344e-03,
                    -7.30173070e-03,
                ],
                [
                    -1.01969471e01,
                    1.95136385e-03,
                    7.65868344e-03,
                    2.78273684e-01,
                    -6.89402590e-02,
                ],
                [
                    2.74302006e00,
                    3.70052723e-03,
                    -7.30173070e-03,
                    -6.89402590e-02,
                    7.12034037e-02,
                ],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        hth = np.array(
            [
                [
                    4.90000000e01,
                    7.04371999e02,
                    1.72131237e03,
                    7.24743592e02,
                    1.70735413e03,
                ],
                [
                    7.04371999e02,
                    1.16866734e04,
                    2.15575320e04,
                    1.10925200e04,
                    2.23848036e04,
                ],
                [
                    1.72131237e03,
                    2.15575320e04,
                    7.39058986e04,
                    2.34796298e04,
                    6.70145378e04,
                ],
                [
                    7.24743592e02,
                    1.10925200e04,
                    2.34796298e04,
                    1.16146226e04,
                    2.30304624e04,
                ],
                [
                    1.70735413e03,
                    2.23848036e04,
                    6.70145378e04,
                    2.30304624e04,
                    6.69879858e04,
                ],
            ]
        )
        np.testing.assert_allclose(reg.hth, hth, RTOL)


class TestGMComboHet(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("CRIME"))
        self.X = np.array(X).T
        self.X = sparse.csr_matrix(self.X)
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        # Only spatial lag
        reg = HET.GM_Combo_Het(self.y, self.X, w=self.w, step1c=True)
        betas = np.array(
            [[57.7778574], [0.73034922], [-0.59257362], [-0.2230231], [0.56636724]]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([25.65156033])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        ef = np.array([31.87664403])
        np.testing.assert_allclose(reg.e_filtered[0], ef, RTOL)
        ep = np.array([28.30648145])
        np.testing.assert_allclose(reg.e_pred[0], ep, RTOL)
        pe = np.array([52.16052155])
        np.testing.assert_allclose(reg.predy_e[0], pe, RTOL)
        predy = np.array([54.81544267])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n)
        k = 4
        np.testing.assert_allclose(reg.k, k)
        y = np.array([80.467003])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.x[0].toarray()[0], x, RTOL)
        yend = np.array([35.4585005])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([18.594, 24.7142675])
        np.testing.assert_allclose(reg.q[0].toarray()[0], q, RTOL)
        z = np.array([1.0, 19.531, 15.72598, 35.4585005])
        np.testing.assert_allclose(reg.z[0].toarray()[0], z, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        stdy = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, stdy)
        vm = np.array(
            [
                [
                    4.86218274e02,
                    -2.77268729e00,
                    -1.59987770e00,
                    -1.01969471e01,
                    2.74302006e00,
                ],
                [
                    -2.77268729e00,
                    1.04680972e-01,
                    2.51172238e-02,
                    1.95136385e-03,
                    3.70052723e-03,
                ],
                [
                    -1.59987770e00,
                    2.51172238e-02,
                    2.15655720e-02,
                    7.65868344e-03,
                    -7.30173070e-03,
                ],
                [
                    -1.01969471e01,
                    1.95136385e-03,
                    7.65868344e-03,
                    2.78273684e-01,
                    -6.89402590e-02,
                ],
                [
                    2.74302006e00,
                    3.70052723e-03,
                    -7.30173070e-03,
                    -6.89402590e-02,
                    7.12034037e-02,
                ],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        pr2 = 0.3001582877472412
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        pr2_e = 0.35613102283621967
        np.testing.assert_allclose(reg.pr2_e, pr2_e, RTOL)
        std_err = np.array(
            [22.05035768, 0.32354439, 0.14685221, 0.52751653, 0.26683966]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        z_stat = np.array(
            [
                (2.6202684885795335, 0.00878605635338265),
                (2.2573385444145524, 0.023986928627746887),
                (-4.0351698589183433, 5.456281036278686e-05),
                (-0.42277935292121521, 0.67245625315942159),
                (2.1225002455741895, 0.033795752094112265),
            ]
        )
        np.testing.assert_allclose(reg.z_stat, z_stat, RTOL)
        hth = np.array(
            [
                [
                    4.90000000e01,
                    7.04371999e02,
                    1.72131237e03,
                    7.24743592e02,
                    1.70735413e03,
                ],
                [
                    7.04371999e02,
                    1.16866734e04,
                    2.15575320e04,
                    1.10925200e04,
                    2.23848036e04,
                ],
                [
                    1.72131237e03,
                    2.15575320e04,
                    7.39058986e04,
                    2.34796298e04,
                    6.70145378e04,
                ],
                [
                    7.24743592e02,
                    1.10925200e04,
                    2.34796298e04,
                    1.16146226e04,
                    2.30304624e04,
                ],
                [
                    1.70735413e03,
                    2.23848036e04,
                    6.70145378e04,
                    2.30304624e04,
                    6.69879858e04,
                ],
            ]
        )
        np.testing.assert_allclose(reg.hth, hth, RTOL)


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
import numpy as np
from spreg import probit as PB
from libpysal.common import RTOL


class TestBaseProbit(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("CRIME"))
        y = np.reshape(y, (49, 1))
        self.y = (y > 40).astype(float)
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("HOVAL"))
        self.X = np.array(X).T
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = PB.BaseProbit(self.y, self.X, w=self.w)
        betas = np.array([[3.35381078], [-0.1996531], [-0.02951371]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        predy = np.array([0.00174739])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 3
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([0.0])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531, 80.467003])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        vm = np.array(
            [
                [8.52813879e-01, -4.36272459e-02, -8.05171472e-03],
                [-4.36272459e-02, 4.11381444e-03, -1.92834842e-04],
                [-8.05171472e-03, -1.92834842e-04, 3.09660240e-04],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        xmean = np.array([[1.0], [14.37493876], [38.43622447]])
        np.testing.assert_allclose(reg.xmean, xmean, RTOL)
        predpc = 85.714285714285708
        np.testing.assert_allclose(reg.predpc, predpc, RTOL)
        logl = -20.06009093055782
        np.testing.assert_allclose(reg.logl, logl, RTOL)
        scale = 0.23309310130643665
        np.testing.assert_allclose(reg.scale, scale, RTOL)
        slopes = np.array([[-0.04653776], [-0.00687944]])
        np.testing.assert_allclose(reg.slopes, slopes, RTOL)
        slopes_vm = np.array(
            [[1.77101993e-04, -1.65021168e-05], [-1.65021168e-05, 1.60575016e-05]]
        )
        np.testing.assert_allclose(reg.slopes_vm, slopes_vm, RTOL)
        LR = 25.317683245671716
        np.testing.assert_allclose(reg.LR[0], LR, RTOL)
        Pinkse_error = 2.9632385352516728
        np.testing.assert_allclose(reg.Pinkse_error[0], Pinkse_error, RTOL)
        KP_error = 1.6509224700582124
        np.testing.assert_allclose(reg.KP_error[0], KP_error, RTOL)
        PS_error = 2.3732463777623511
        np.testing.assert_allclose(reg.PS_error[0], PS_error, RTOL)


class TestProbit(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("CRIME"))
        y = np.reshape(y, (49, 1))
        self.y = (y > 40).astype(float)
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("HOVAL"))
        self.X = np.array(X).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = PB.Probit(self.y, self.X, w=self.w)
        betas = np.array([[3.35381078], [-0.1996531], [-0.02951371]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        predy = np.array([0.00174739])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 3
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([0.0])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 19.531, 80.467003])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        vm = np.array(
            [
                [8.52813879e-01, -4.36272459e-02, -8.05171472e-03],
                [-4.36272459e-02, 4.11381444e-03, -1.92834842e-04],
                [-8.05171472e-03, -1.92834842e-04, 3.09660240e-04],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        xmean = np.array([[1.0], [14.37493876], [38.43622447]])
        np.testing.assert_allclose(reg.xmean, xmean, RTOL)
        predpc = 85.714285714285708
        np.testing.assert_allclose(reg.predpc, predpc, RTOL)
        logl = -20.06009093055782
        np.testing.assert_allclose(reg.logl, logl, RTOL)
        scale = 0.23309310130643665
        np.testing.assert_allclose(reg.scale, scale, RTOL)
        slopes = np.array([[-0.04653776], [-0.00687944]])
        np.testing.assert_allclose(reg.slopes, slopes, RTOL)
        slopes_vm = np.array(
            [[1.77101993e-04, -1.65021168e-05], [-1.65021168e-05, 1.60575016e-05]]
        )
        np.testing.assert_allclose(reg.slopes_vm, slopes_vm, RTOL)
        LR = 25.317683245671716
        np.testing.assert_allclose(reg.LR[0], LR, RTOL)
        Pinkse_error = 2.9632385352516728
        np.testing.assert_allclose(reg.Pinkse_error[0], Pinkse_error, RTOL)
        KP_error = 1.6509224700582124
        np.testing.assert_allclose(reg.KP_error[0], KP_error, RTOL)
        PS_error = 2.3732463777623511
        np.testing.assert_allclose(reg.PS_error[0], PS_error, RTOL)


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import libpysal
from libpysal.common import RTOL
from spreg.sp_panels import *

ATOL = 1e-12


class Test_GM_KKP(unittest.TestCase):
    def setUp(self):
        nat = libpysal.examples.load_example("NCOVR")
        self.db = libpysal.io.open(nat.get_path("NAT.dbf"), "r")
        self.w = libpysal.weights.Queen.from_shapefile(
            libpysal.examples.get_path("NAT.shp")
        )
        self.w.transform = "r"
        self.y_var0 = ["HR70", "HR80", "HR90"]
        self.x_var0 = ["RD70", "RD80", "RD90", "PS70", "PS80", "PS90"]
        self.y = np.array([self.db.by_col(name) for name in self.y_var0]).T
        self.x = np.array([self.db.by_col(name) for name in self.x_var0]).T

    def test_wide_ident(self):
        reg = GM_KKP(
            self.y,
            self.x,
            self.w,
            full_weights=False,
            name_y=self.y_var0,
            name_x=self.x_var0,
        )
        np.testing.assert_allclose(
            reg.betas,
            np.array(
                [
                    [6.49221562],
                    [3.62445753],
                    [1.31187779],
                    [0.41777589],
                    [22.81908224],
                    [39.90993228],
                ]
            ),
            RTOL,
        )
        np.testing.assert_allclose(
            reg.vm,
            np.array(
                [
                    [1.26948117e-02, -1.98160325e-06, 7.38157674e-05],
                    [-1.98160325e-06, 7.69961725e-03, 1.13099329e-03],
                    [7.38157674e-05, 1.13099329e-03, 7.26783636e-03],
                ]
            ),
            RTOL,
        )
        np.testing.assert_equal(
            reg.name_x, ["CONSTANT", "RD", "PS", "lambda", " sigma2_v", "sigma2_1"]
        )
        np.testing.assert_equal(reg.name_y, "HR")

    def test_wide_full(self):
        reg = GM_KKP(self.y, self.x, self.w, full_weights=True)

        np.testing.assert_allclose(
            reg.betas,
            np.array(
                [
                    [6.49193589],
                    [3.55740165],
                    [1.29462748],
                    [0.4263399],
                    [22.47241979],
                    [45.82593532],
                ]
            ),
            RTOL,
        )
        np.testing.assert_allclose(
            reg.vm,
            np.array(
                [
                    [1.45113773e-02, -2.14882672e-06, 8.54997693e-05],
                    [-2.14882672e-06, 8.41929187e-03, 1.24553497e-03],
                    [8.54997693e-05, 1.24553497e-03, 8.12448812e-03],
                ]
            ),
            RTOL,
        )

    def test_long_ident(self):
        bigy = self.y.reshape((self.y.size, 1), order="F")
        bigx = self.x[:, 0:3].reshape((self.x.shape[0] * 3, 1), order="F")
        bigx = np.hstack(
            (bigx, self.x[:, 3:6].reshape((self.x.shape[0] * 3, 1), order="F"))
        )
        reg = GM_KKP(
            bigy, bigx, self.w, full_weights=False, name_y=["HR"], name_x=["RD", "PS"]
        )

        np.testing.assert_allclose(
            reg.betas,
            np.array(
                [
                    [6.49221562],
                    [3.62445753],
                    [1.31187779],
                    [0.41777589],
                    [22.81908224],
                    [39.90993228],
                ]
            ),
            RTOL,
        )
        np.testing.assert_allclose(
            reg.vm,
            np.array(
                [
                    [1.26948117e-02, -1.98160325e-06, 7.38157674e-05],
                    [-1.98160325e-06, 7.69961725e-03, 1.13099329e-03],
                    [7.38157674e-05, 1.13099329e-03, 7.26783636e-03],
                ]
            ),
            RTOL,
        )
        np.testing.assert_equal(
            reg.name_x, ["CONSTANT", "RD", "PS", "lambda", " sigma2_v", "sigma2_1"]
        )
        np.testing.assert_equal(reg.name_y, "HR")

    def test_regimes(self):
        regimes = self.db.by_col("SOUTH")
        reg = GM_KKP(
            self.y,
            self.x,
            self.w,
            full_weights=False,
            regimes=regimes,
            name_y=self.y_var0,
            name_x=self.x_var0,
        )
        np.testing.assert_allclose(
            reg.betas,
            np.array(
                [
                    [5.25856482],
                    [3.19249165],
                    [1.0056967],
                    [7.94560642],
                    [3.13931041],
                    [1.53700634],
                    [0.35979407],
                    [22.5650005],
                    [39.71516708],
                ]
            ),
            RTOL,
        )
        np.testing.assert_allclose(
            np.sqrt(reg.vm.diagonal()),
            np.array([0.158986, 0.157543, 0.104128, 0.165254, 0.117737, 0.136666]),
            RTOL,
        )
        np.testing.assert_equal(
            reg.name_x,
            [
                "0_CONSTANT",
                "0_RD",
                "0_PS",
                "1_CONSTANT",
                "1_RD",
                "1_PS",
                "lambda",
                " sigma2_v",
                "sigma2_1",
            ],
        )
        np.testing.assert_equal(reg.name_y, "HR")
        np.testing.assert_allclose(
            reg.chow.regi,
            np.array(
                [
                    [1.420430e02, 9.516507e-33],
                    [7.311490e-02, 7.868543e-01],
                    [9.652492e00, 1.890949e-03],
                ]
            ),
            RTOL,
        )
        np.testing.assert_allclose(reg.chow.joint[0], 158.7225, RTOL)


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
from scipy import sparse
import numpy as np
from spreg.ml_lag import ML_Lag
from libpysal.common import RTOL
from warnings import filterwarnings

filterwarnings("ignore", category=sparse.SparseEfficiencyWarning)


class TestMLError(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("baltim.dbf"), "r")
        self.ds_name = "baltim.dbf"
        self.y_name = "PRICE"
        self.y = np.array(db.by_col(self.y_name)).T
        self.y.shape = (len(self.y), 1)
        self.x_names = ["NROOM", "AGE", "SQFT"]
        self.x = np.array([db.by_col(var) for var in self.x_names]).T
        ww = libpysal.io.open(libpysal.examples.get_path("baltim_q.gal"))
        self.w = ww.read()
        ww.close()
        self.w_name = "baltim_q.gal"
        self.w.transform = "r"

    def _estimate_and_compare(self, **kwargs):
        reg = ML_Lag(
            self.y,
            self.x,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w=self.w_name,
            name_ds=self.ds_name,
            **kwargs
        )
        betas = np.array(
            [[-6.04040164], [3.48995114], [-0.20103955], [0.65462382], [0.62351143]]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([47.51218398])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([-0.51218398])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 211
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 5
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([47.0])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 4.0, 148.0, 11.25])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        e = np.array([41.99251608])
        np.testing.assert_allclose(reg.e_pred[0], e, RTOL)
        my = 44.307180094786695
        np.testing.assert_allclose(reg.mean_y, my)
        sy = 23.606076835380495
        np.testing.assert_allclose(reg.std_y, sy)
        vm = np.array([28.57288755, 1.42341656, 0.00288068, 0.02956392, 0.00332139])
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = 216.27525647243797
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.6133020721559487
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array([5.34536131, 1.19307022, 0.05367198, 0.17194162, 0.05763147])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        logll = -875.92771143484833
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 1761.8554228696967
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 1778.614713537077
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)

    def test_dense(self):
        self._estimate_and_compare(method="FULL")

    def test_ord(self):
        reg = ML_Lag(
            self.y,
            self.x,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w=self.w_name,
            name_ds=self.ds_name,
            method="ORD",
        )
        betas = np.array(
            [[-6.04040164], [3.48995114], [-0.20103955], [0.65462382], [0.62351143]]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([47.51218398])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([-0.51218398])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 211
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 5
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([47.0])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 4.0, 148.0, 11.25])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        e = np.array([41.99251608])
        np.testing.assert_allclose(reg.e_pred[0], e, RTOL)
        my = 44.307180094786695
        np.testing.assert_allclose(reg.mean_y, my)
        sy = 23.606076835380495
        np.testing.assert_allclose(reg.std_y, sy)
        vm = np.array([28.63404, 1.423698, 0.002884738, 0.02957845, 0.003379166])
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = 216.27525647243797
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.6133020721559487
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array([5.351078, 1.193188, 0.05371, 0.171984, 0.058131])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        logll = -875.92771143484833
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 1761.8554228696967
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 1778.614713537077
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)

    def test_LU(self):
        self._estimate_and_compare(method="LU")


if __name__ == "__main__":
    unittest.main()



import unittest
import scipy
import libpysal
import numpy as np
from spreg import error_sp_regimes as SP
from spreg.error_sp import GM_Error, GM_Endog_Error, GM_Combo
from libpysal.common import RTOL


class TestGM_Error_Regimes(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("CRIME"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("HOVAL"))
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        self.w = libpysal.weights.Queen.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"
        self.r_var = "NSA"
        self.regimes = db.by_col(self.r_var)
        X1 = []
        X1.append(db.by_col("INC"))
        self.X1 = np.array(X1).T
        yd = []
        yd.append(db.by_col("HOVAL"))
        self.yd = np.array(yd).T
        q = []
        q.append(db.by_col("DISCBD"))
        self.q = np.array(q).T
        # Artficial:
        n = 256
        self.n2 = n // 2
        self.x_a1 = np.random.uniform(-10, 10, (n, 1))
        self.x_a2 = np.random.uniform(1, 5, (n, 1))
        self.q_a = self.x_a2 + np.random.normal(0, 1, (n, 1))
        self.x_a = np.hstack((self.x_a1, self.x_a2))
        self.y_a = np.dot(
            np.hstack((np.ones((n, 1)), self.x_a)), np.array([[1], [0.5], [2]])
        ) + np.random.normal(0, 1, (n, 1))
        latt = int(np.sqrt(n))
        self.w_a = libpysal.weights.util.lat2W(latt, latt)
        self.w_a.transform = "r"
        self.regi_a = [0] * (n // 2) + [1] * (n // 2)  ##must be floors!
        self.w_a1 = libpysal.weights.util.lat2W(latt // 2, latt)
        self.w_a1.transform = "r"

    def test_model(self):
        reg = SP.GM_Error_Regimes(self.y, self.X, self.regimes, self.w)
        betas = np.array(
            [
                [63.3443073],
                [-0.15468],
                [-1.52186509],
                [61.40071412],
                [-0.33550084],
                [-0.85076108],
                [0.38671608],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([-2.06177251])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([17.78775251])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 6
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 0.0, 1.0, 80.467003, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        e = np.array([1.40747232])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array([50.55875289, -0.14444487, -2.05735489, 0.0, 0.0, 0.0])
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        sig2 = 102.13050615267227
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.5525102200608539
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        std_err = np.array(
            [7.11046784, 0.21879293, 0.58477864, 7.50596504, 0.10800686, 0.57365981]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [
                [0.03533785, 0.85088948],
                [0.54918491, 0.45865093],
                [0.67115641, 0.41264872],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.81985446000130979
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    """
    def test_model_regi_error(self):
        #Columbus:
        reg = SP.GM_Error_Regimes(self.y, self.X, self.regimes, self.w, regime_err_sep=True)
        betas = np.array([[ 60.45730439],
       [ -0.17732134],
       [ -1.30936328],
       [  0.51314713],
       [ 66.5487126 ],
       [ -0.31845995],
       [ -1.29047149],
       [  0.08092997]])
        np.testing.assert_allclose(reg.betas,betas,RTOL)
        vm = np.array([ 39.33656288,  -0.08420799,  -1.50350999,   0.        ,
         0.        ,   0.        ])
        np.testing.assert_allclose(reg.vm[0],vm,RTOL)
        u = np.array([ 0.00698341])
        np.testing.assert_allclose(reg.u[0],u,RTOL)
        predy = np.array([ 15.71899659])
        np.testing.assert_allclose(reg.predy[0],predy,RTOL)
        e = np.array([ 0.53685671])
        np.testing.assert_allclose(reg.e_filtered[0],e,RTOL)
        chow_r = np.array([[  3.63674458e-01,   5.46472584e-01],
       [  4.29607250e-01,   5.12181727e-01],
       [  5.44739543e-04,   9.81379339e-01]])
        np.testing.assert_allclose(reg.chow.regi,chow_r,RTOL)
        chow_j = 0.70119418251625387
        np.testing.assert_allclose(reg.chow.joint[0],chow_j,RTOL)
        #Artficial:
        model = SP.GM_Error_Regimes(self.y_a, self.x_a, self.regi_a, w=self.w_a, regime_err_sep=True)
        model1 = GM_Error(self.y_a[0:(self.n2)].reshape((self.n2),1), self.x_a[0:(self.n2)], w=self.w_a1)
        model2 = GM_Error(self.y_a[(self.n2):].reshape((self.n2),1), self.x_a[(self.n2):], w=self.w_a1)
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas,tbetas,RTOL)
        vm = np.hstack((model1.vm.diagonal(),model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm,RTOL)
    """

    def test_model_endog(self):
        reg = SP.GM_Endog_Error_Regimes(
            self.y, self.X1, self.yd, self.q, self.regimes, self.w
        )
        betas = np.array(
            [
                [
                    77.48385551,
                    4.52986622,
                    78.93209405,
                    0.42186261,
                    -3.23823854,
                    -1.1475775,
                    0.20222108,
                ]
            ]
        )
        print("Runining higher-tolerance test on L133 of test_error_sp_regimes.py")
        np.testing.assert_allclose(reg.betas.T, betas, RTOL + 0.0001)
        u = np.array([20.89660904])
        # np.testing.assert_allclose(reg.u[0],u,RTOL)
        np.testing.assert_allclose(reg.u[0], u, rtol=1e-05)
        e = np.array([25.21818724])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        predy = np.array([-5.17062904])
        # np.testing.assert_allclose(reg.predy[0],predy,RTOL)
        np.testing.assert_allclose(reg.predy[0], predy, rtol=1e-03)
        n = 49
        np.testing.assert_allclose(reg.n, n)
        k = 6
        np.testing.assert_allclose(reg.k, k)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 1.0, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        yend = np.array([[0.0, 80.467003]])
        np.testing.assert_allclose(reg.yend[0].toarray(), yend, RTOL)
        z = np.array([[0.0, 0.0, 1.0, 19.531, 0.0, 80.467003]])
        np.testing.assert_allclose(reg.z[0].toarray(), z, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy)
        vm = np.array([390.88250241, 52.25924084, 0.0, 0.0, -32.64274729, 0.0])
        # np.testing.assert_allclose(reg.vm[0],vm,RTOL)
        np.allclose(reg.vm, vm)
        pr2 = 0.19623994206233333
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        sig2 = 649.4011
        # np.testing.assert_allclose(round(reg.sig2,RTOL),round(sig2,RTOL),RTOL)
        np.testing.assert_allclose(sig2, reg.sig2, rtol=1e-05)
        std_err = np.array(
            [19.77074866, 6.07667394, 24.32254786, 2.17776972, 2.97078606, 0.94392418]
        )
        # np.testing.assert_allclose(reg.std_err,std_err,RTOL)
        np.testing.assert_allclose(reg.std_err, std_err, rtol=1e-05)
        chow_r = np.array(
            [[0.0021348, 0.96314775], [0.40499741, 0.5245196], [0.4498365, 0.50241261]]
        )
        print("Running higher-tolerance tests on L176 of test_error_sp_regimes.py")
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL + 0.0001)
        chow_j = 1.2885590185243503
        # np.testing.assert_allclose(reg.chow.joint[0],chow_j)
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, rtol=1e-05)

    def test_model_endog_regi_error(self):
        # Columbus:
        reg = SP.GM_Endog_Error_Regimes(
            self.y, self.X1, self.yd, self.q, self.regimes, self.w, regime_err_sep=True
        )
        betas = np.array(
            [
                [7.91729500e01],
                [5.80693176e00],
                [-3.84036576e00],
                [1.46462983e-01],
                [8.24723791e01],
                [5.68908920e-01],
                [-1.28824699e00],
                [6.70387351e-02],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array([791.86679123, 140.12967794, -81.37581255, 0.0, 0.0, 0.0])
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([25.80361497])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([-10.07763497])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([27.32251813])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [0.00926459, 0.92331985],
                [0.26102777, 0.60941494],
                [0.26664581, 0.60559072],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 1.1184631131987004
        # np.testing.assert_allclose(reg.chow.joint[0],chow_j)
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = SP.GM_Endog_Error_Regimes(
            self.y_a,
            self.x_a1,
            yend=self.x_a2,
            q=self.q_a,
            regimes=self.regi_a,
            w=self.w_a,
            regime_err_sep=True,
        )
        model1 = GM_Endog_Error(
            self.y_a[0 : (self.n2)].reshape((self.n2), 1),
            self.x_a1[0 : (self.n2)],
            yend=self.x_a2[0 : (self.n2)],
            q=self.q_a[0 : (self.n2)],
            w=self.w_a1,
        )
        model2 = GM_Endog_Error(
            self.y_a[(self.n2) :].reshape((self.n2), 1),
            self.x_a1[(self.n2) :],
            yend=self.x_a2[(self.n2) :],
            q=self.q_a[(self.n2) :],
            w=self.w_a1,
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas, RTOL)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm, RTOL)

    def test_model_combo(self):
        reg = SP.GM_Combo_Regimes(
            self.y, self.X1, self.regimes, self.yd, self.q, w=self.w
        )
        predy_e = np.array([18.82774339])
        np.testing.assert_allclose(reg.predy_e[0], predy_e, RTOL)
        betas = np.array(
            [
                [36.44798052],
                [-0.7974482],
                [30.53782661],
                [-0.72602806],
                [-0.30953121],
                [-0.21736652],
                [0.64801059],
                [-0.16601265],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([0.84393304])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        e_filtered = np.array([0.4040027])
        np.testing.assert_allclose(reg.e_filtered[0], e_filtered, RTOL)
        predy = np.array([14.88204696])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 7
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 1.0, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        yend = np.array([[0.0, 80.467003, 24.7142675]])
        np.testing.assert_allclose(reg.yend[0].toarray(), yend, RTOL)
        z = np.array([[0.0, 0.0, 1.0, 19.531, 0.0, 80.467003, 24.7142675]])
        np.testing.assert_allclose(reg.z[0].toarray(), z, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                109.23549239,
                -0.19754121,
                84.29574673,
                -1.99317178,
                -1.60123994,
                -0.1252719,
                -1.3930344,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        sig2 = 94.98610921110007
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.6493586702255537
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        pr2_e = 0.5255332447240576
        np.testing.assert_allclose(reg.pr2_e, pr2_e, RTOL)
        std_err = np.array(
            [
                10.45157846,
                0.93942923,
                11.38484969,
                0.60774708,
                0.44461334,
                0.15871227,
                0.15738141,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [
                [0.49716076, 0.48075032],
                [0.00405377, 0.94923363],
                [0.03866684, 0.84411016],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.64531386285872072
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_combo_regi_error(self):
        # Columbus:
        reg = SP.GM_Combo_Regimes(
            self.y,
            self.X1,
            self.regimes,
            self.yd,
            self.q,
            w=self.w,
            regime_lag_sep=True,
            regime_err_sep=True,
        )
        betas = np.array(
            [
                [42.01035248],
                [-0.13938772],
                [-0.6528306],
                [0.54737621],
                [0.2684419],
                [34.02473255],
                [-0.14920259],
                [-0.48972903],
                [0.65883658],
                [-0.17174845],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array(
            [153.58614432, 2.96302131, -3.26211855, -2.46914703, 0.0, 0.0, 0.0, 0.0]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([7.73968703])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([7.98629297])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([6.45052714])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [1.00886404e-01, 7.50768497e-01],
                [3.61843271e-05, 9.95200481e-01],
                [4.69585772e-02, 8.28442711e-01],
                [8.13275259e-02, 7.75506385e-01],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.28479988992843119
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = SP.GM_Combo_Regimes(
            self.y_a,
            self.x_a1,
            yend=self.x_a2,
            q=self.q_a,
            regimes=self.regi_a,
            w=self.w_a,
            regime_err_sep=True,
            regime_lag_sep=True,
        )
        model1 = GM_Combo(
            self.y_a[0 : (self.n2)].reshape((self.n2), 1),
            self.x_a1[0 : (self.n2)],
            yend=self.x_a2[0 : (self.n2)],
            q=self.q_a[0 : (self.n2)],
            w=self.w_a1,
        )
        model2 = GM_Combo(
            self.y_a[(self.n2) :].reshape((self.n2), 1),
            self.x_a1[(self.n2) :],
            yend=self.x_a2[(self.n2) :],
            q=self.q_a[(self.n2) :],
            w=self.w_a1,
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import libpysal
from spreg import diagnostics
from spreg.ols import OLS
from libpysal.common import RTOL

# create regression object used by all the tests below
db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"))
y = np.array(db.by_col("CRIME"))
y = np.reshape(y, (49, 1))
X = []
X.append(db.by_col("INC"))
X.append(db.by_col("HOVAL"))
X = np.array(X).T
reg = OLS(y, X)


class TestFStat(unittest.TestCase):
    def test_f_stat(self):
        obs = diagnostics.f_stat(reg)
        exp = (28.385629224695, 0.000000009340747)
        for i in range(2):
            np.testing.assert_allclose(obs[i], exp[i], RTOL)


class TestTStat(unittest.TestCase):
    def test_t_stat(self):
        obs = diagnostics.t_stat(reg)
        exp = [
            (14.490373143689094, 9.2108899889173982e-19),
            (-4.7804961912965762, 1.8289595070843232e-05),
            (-2.6544086427176916, 0.010874504909754612),
        ]
        for i in range(3):
            for j in range(2):
                np.testing.assert_allclose(obs[i][j], exp[i][j], RTOL)


class TestR2(unittest.TestCase):
    def test_r2(self):
        obs = diagnostics.r2(reg)
        exp = 0.55240404083742334
        np.testing.assert_allclose(obs, exp, RTOL)


class TestAr2(unittest.TestCase):
    def test_ar2(self):
        obs = diagnostics.ar2(reg)
        exp = 0.5329433469607896
        np.testing.assert_allclose(obs, exp, RTOL)


class TestSeBetas(unittest.TestCase):
    def test_se_betas(self):
        obs = diagnostics.se_betas(reg)
        exp = np.array([4.73548613, 0.33413076, 0.10319868])
        np.testing.assert_allclose(obs, exp, RTOL)


class TestLogLikelihood(unittest.TestCase):
    def test_log_likelihood(self):
        obs = diagnostics.log_likelihood(reg)
        exp = -187.3772388121491
        np.testing.assert_allclose(obs, exp, RTOL)


class TestAkaike(unittest.TestCase):
    def test_akaike(self):
        obs = diagnostics.akaike(reg)
        exp = 380.7544776242982
        np.testing.assert_allclose(obs, exp, RTOL)


class TestSchwarz(unittest.TestCase):
    def test_schwarz(self):
        obs = diagnostics.schwarz(reg)
        exp = 386.42993851863008
        np.testing.assert_allclose(obs, exp, RTOL)


class TestConditionIndex(unittest.TestCase):
    def test_condition_index(self):
        obs = diagnostics.condition_index(reg)
        exp = 6.541827751444
        np.testing.assert_allclose(obs, exp, RTOL)


class TestJarqueBera(unittest.TestCase):
    def test_jarque_bera(self):
        obs = diagnostics.jarque_bera(reg)
        exp = {"df": 2, "jb": 1.835752520076, "pvalue": 0.399366291249}
        np.testing.assert_equal(obs["df"], exp["df"])
        np.testing.assert_allclose(obs["jb"], exp["jb"], RTOL)
        np.testing.assert_allclose(obs["pvalue"], exp["pvalue"], RTOL)


class TestBreuschPagan(unittest.TestCase):
    def test_breusch_pagan(self):
        obs = diagnostics.breusch_pagan(reg)
        exp = {"df": 2, "bp": 7.900441675960, "pvalue": 0.019250450075}
        np.testing.assert_equal(obs["df"], exp["df"])
        np.testing.assert_allclose(obs["bp"], exp["bp"])
        np.testing.assert_allclose(obs["pvalue"], exp["pvalue"])


class TestWhite(unittest.TestCase):
    def test_white(self):
        obs = diagnostics.white(reg)
        exp = {"df": 5, "wh": 19.946008239903, "pvalue": 0.001279222817}
        np.testing.assert_equal(obs["df"], exp["df"])
        np.testing.assert_allclose(obs["wh"], exp["wh"], RTOL)
        np.testing.assert_allclose(obs["pvalue"], exp["pvalue"], RTOL)


class TestKoenkerBassett(unittest.TestCase):
    def test_koenker_bassett(self):
        obs = diagnostics.koenker_bassett(reg)
        exp = {"df": 2, "kb": 5.694087931707, "pvalue": 0.058015563638}
        np.testing.assert_equal(obs["df"], exp["df"])
        np.testing.assert_allclose(obs["kb"], exp["kb"], RTOL)
        np.testing.assert_allclose(obs["pvalue"], exp["pvalue"], RTOL)


class TestVif(unittest.TestCase):
    def test_vif(self):
        obs = diagnostics.vif(reg)
        exp = [
            (0.0, 0.0),  # note [0][1] should actually be infiniity...
            (1.3331174971891975, 0.75012142748740696),
            (1.3331174971891973, 0.75012142748740707),
        ]
        for i in range(1, 3):
            for j in range(2):
                np.testing.assert_allclose(obs[i][j], exp[i][j], RTOL)


class TestConstantCheck(unittest.TestCase):
    def test_constant_check(self):
        obs = diagnostics.constant_check(reg.x)
        exp = True
        np.testing.assert_equal(obs, exp)


if __name__ == "__main__":
    unittest.main()



"""Skater Regression classes."""

__author__ = "Luc Anselin anselin@uchicago.edu, Pedro Amaral pedroamaral@cedeplar.ufmg.br, Levi Wolf levi.john.wolf@bristol.ac.uk"

from scipy.sparse import csgraph as cg
from scipy.optimize import OptimizeWarning
from scipy.spatial.distance import cdist
from collections import namedtuple
from warnings import warn
from libpysal.weights import w_subset
from .utils import set_endog
import time
import numpy as np
import copy

try:
    from sklearn.metrics import euclidean_distances
except ImportError:
    from scipy.spatial.distance import pdist, cdist, squareform

    def euclidean_distances(X, Y=None):
        """
        fallback function to compute pairwise euclidean distances
        for a single input, or point-to-point euclidean distances
        for two inputs.
        """
        if Y is None:
            return squareform(pdist(X))
        else:
            return cdist(X, Y)


__all__ = ["Skater_reg"]

deletion = namedtuple("deletion", ("in_node", "out_node", "score"))


class Skater_reg(object):
    """
    Initialize the Skater_reg algorithm based on :cite:`Anselin2021`.
    The function can currently estimate OLS, from
    spreg or stats_models, and Spatial Lag models from spreg.
    Fit method performs estimation and returns a Skater_reg object.

    Parameters
    ----------
    dissimilarity : a callable distance metric.
                    Default: sklearn.metrics.pairwise.euclidean_distances
    affinity      : a callable affinity metric between 0,1.
                    Will be inverted to provide a
                    dissimilarity metric.
    reduction     : the reduction applied over all clusters
                    to provide the map score.
                    Default: numpy.sum
    center        : way to compute the center of each region in attribute space
                    Default: numpy.mean

    NOTE: Optimization occurs with respect to a *dissimilarity* metric, so the reduction should
              yield some kind of score where larger values are *less desirable* than smaller values.
              Typically, this means we use addition.


    Attributes
    ----------
    coords        : array-like
                    n*2, collection of n sets of (x,y) coordinates used for
                    calibration locations
    y             : array
                    n*1, dependent variable
    X             : array
                    n*k, independent variable, exlcuding the constant
    bw            : scalar
                    bandwidth value consisting of either a distance or N
                    nearest neighbors; user specified or obtained using
                    Sel_BW
    family        : family object
                    underlying probability model; provides
                    distribution-specific calculations
    offset        : array
                    n*1, the offset variable at the ith location. For Poisson model
                    this term is often the size of the population at risk or
                    the expected size of the outcome in spatial epidemiology
                    Default is None where Ni becomes 1.0 for all locations
    sigma2_v1     : boolean
                    specify form of corrected denominator of sigma squared to use for
                    model diagnostics; Acceptable options are:
                    'True':       n-tr(S) (defualt)
                    'False':     n-2(tr(S)+tr(S'S))
    kernel        : string
                    type of kernel function used to weight observations;
                    available options:
                    'gaussian'
                    'bisquare'
                    'exponential'
    fixed         : boolean
                    True for distance based kernel function and  False for
                    adaptive (nearest neighbor) kernel function (default)
    constant      : boolean
                    True to include intercept (default) in model and False to exclude
                    intercept
    spherical     : boolean
                    True for shperical coordinates (long-lat),
                    False for projected coordinates (defalut).
    hat_matrix    : boolean
                    True to store full n by n hat matrix,
                    False to not store full hat matrix to minimize memory footprint (defalut).
    n             : integer
                    number of observations
    k             : integer
                    number of independent variables
    mean_y        : float
                    mean of y
    std_y         : float
                    standard deviation of y
    fit_params    : dict
                    parameters passed into fit method to define estimation
                    routine
    points        : array-like
                    n*2, collection of n sets of (x,y) coordinates used for
                    calibration locations instead of all observations;
                    defaults to None unles specified in predict method
    P             : array
                    n*k, independent variables used to make prediction;
                    exlcuding the constant; default to None unless specified
                    in predict method
    exog_scale    : scalar
                    estimated scale using sampled locations; defualt is None
                    unless specified in predict method
    exog_resid    : array-like
                    estimated residuals using sampled locations; defualt is None
                    unless specified in predict method
    Examples
    --------
    >>> import libpysal as ps
    >>> import numpy as np
    >>> import spreg
    >>> from spreg.skater_reg import Skater_reg
    >>> data = ps.io.open(ps.examples.get_path('columbus.dbf'))
    >>> y = np.array(data.by_col('HOVAL')).reshape((-1,1))
    >>> x_var = ['INC','CRIME']
    >>> x = np.array([db.by_col(name) for name in x_var]).T
    >>> w = ps.weights.Queen.from_shapefile(ps.examples.get_path("columbus.shp"))
    >>> x_std = (x - np.mean(x,axis=0)) / np.std(x,axis=0)
    >>> results = Skater_reg().fit(3, w, x_std, {'reg':spreg.OLS,'y':y,'x':x}, quorum=10, trace=False)
    >>> results.current_labels_
    array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 1, 0,
       0, 1, 0, 0, 2, 2, 0, 0, 1, 0, 2, 1, 2, 1, 2, 0, 1, 0, 0, 1, 2, 2,
       2, 1, 0, 2, 2], dtype=int32)
    """

    def __init__(
        self,
        dissimilarity=euclidean_distances,
        affinity=None,
        reduction=np.sum,
        center=np.mean,
    ):
        if affinity is not None:
            # invert the 0,1 affinity to
            # to an unbounded positive dissimilarity
            metric = lambda x: -np.log(affinity(x))
        else:
            metric = dissimilarity
        self.metric = metric
        self.reduction = reduction
        self.center = center

    def __repr__(self):
        return "Skater_reg_object(metric = {}, reduction = {}, center = {})".format(
            self.metric, self.reduction, self.center
        )

    def fit(
        self,
        n_clusters,
        W,
        data=None,
        data_reg=None,
        quorum=-np.inf,
        trace=True,
        islands="increase",
        verbose=False,
        model_family="spreg",
    ):
        """
        Method that fits a model with a particular estimation routine.

        Parameters
        ----------
        n_clusters   : int of clusters wanted
        W            : pysal W object expressing the neighbor relationships between observations.
                       Should be symmetric and binary, so Queen/Rook, DistanceBand, or a symmetrized KNN.
        data         : np.ndarray of (N,P) shape with N observations and P features
                       This is the data that is used to evaluate the similarity between each observation.
        data_reg     : list containing:
                       1- a callable regression method (ex. OLS or GM_Lag from spreg or OLS from statsmodels)
                       2- np.ndarray of (N,1) shape with N observations on the depedent variable for the regression
                       3- np.ndarray of (N,k) shape with N observations and k columns containing the explanatory variables (constant must not be included)
                       4- pysal W object to be used in the regression (optional)
        quorum       : int with minimum size of each region.
        trace        : bool denoting whether to store intermediate
                       labelings as the tree gets pruned
        islands      : string describing what to do with islands.
                       If "ignore", will discover `n_clusters` regions, treating islands as their own regions.
                       If "increase", will discover `n_clusters` regions, treating islands as separate from n_clusters.
        verbose      : bool/int describing how much output to provide to the user,
                       in terms of print statements or progressbars.
        model_family : string describing the fFamily of estimation method used for the regression.
                       Must be either 'spreg' (default) or 'statsmodels'

        Returns
        -------
                     : Skater_reg object.
        """
        if trace:
            self._trace = []
        if data is None:
            attribute_kernel = np.ones((W.n, W.n))
            data = np.ones((W.n, 1))
        else:
            attribute_kernel = self.metric(data)
        W.transform = "b"
        W = W.sparse
        start = time.time()

        super_verbose = verbose > 1
        start_W = time.time()
        dissim = W.multiply(attribute_kernel)
        dissim.eliminate_zeros()
        end_W = time.time() - start_W

        if super_verbose:
            print("Computing Affinity Kernel took {:.2f}s".format(end_W))

        tree_time = time.time()
        MSF = cg.minimum_spanning_tree(dissim)
        tree_time = time.time() - tree_time
        if super_verbose:
            print("Computing initial MST took {:.2f}s".format(tree_time))

        initial_component_time = time.time()
        current_n_subtrees, current_labels = cg.connected_components(
            MSF, directed=False
        )
        initial_component_time = time.time() - initial_component_time

        if super_verbose:
            print(
                "Computing connected components took {:.2f}s.".format(
                    initial_component_time
                )
            )

        if current_n_subtrees > 1:
            island_warnings = [
                "Increasing `n_clusters` from {} to {} in order to account for islands.".format(
                    n_clusters, n_clusters + current_n_subtrees
                ),
                "Counting islands towards the remaining {} clusters.".format(
                    n_clusters - (current_n_subtrees)
                ),
            ]
            ignoring_islands = int(islands.lower() == "ignore")
            chosen_warning = island_warnings[ignoring_islands]
            warn(
                "By default, the graph is disconnected! {}".format(chosen_warning),
                OptimizeWarning,
                stacklevel=2,
            )
            if not ignoring_islands:
                n_clusters += current_n_subtrees
            _, island_populations = np.unique(current_labels, return_counts=True)
            if (island_populations < quorum).any():
                raise ValueError(
                    "Islands must be larger than the quorum. If not, drop the small islands and solve for"
                    " clusters in the remaining field."
                )
        if trace:
            self._trace.append(([], deletion(np.nan, np.nan, np.inf)))
            if super_verbose:
                print(self._trace[-1])
        trees_scores = None
        prev_score = np.inf
        while current_n_subtrees < n_clusters:  # while we don't have enough regions
            (
                best_deletion,
                trees_scores,
                new_MSF,
                current_n_subtrees,
                current_labels,
            ) = self.find_cut(
                MSF,
                data,
                data_reg,
                current_n_subtrees,
                current_labels,
                quorum=quorum,
                trees_scores=trees_scores,
                labels=None,
                target_label=None,
                verbose=verbose,
                model_family=model_family,
            )

            if np.isfinite(best_deletion.score):  # if our search succeeds
                # accept the best move as *the* move
                if super_verbose:
                    print("cut made {}...".format(best_deletion))
                if best_deletion.score > prev_score:
                    raise ValueError(
                        "The score increased with the number of clusters. Please check your data."
                    )
                prev_score = best_deletion.score
                MSF = new_MSF
            else:  # otherwise, it means the MSF admits no further cuts
                prev_n_subtrees, _ = cg.connected_components(MSF, directed=False)
                warn(
                    "MSF contains no valid moves after finding {} subtrees. "
                    "Decrease the size of your quorum to find the remaining {} subtrees.".format(
                        prev_n_subtrees, n_clusters - prev_n_subtrees
                    ),
                    OptimizeWarning,
                    stacklevel=2,
                )
            if trace:
                self._trace.append((current_labels, best_deletion))

        self.current_labels_ = current_labels
        self.minimum_spanning_forest_ = MSF
        self._elapsed_time = time.time() - start
        return self

    def score_spreg(
        self,
        data=None,
        data_reg=None,
        all_labels=None,
        quorum=-np.inf,
        current_labels=None,
        current_tree=None,
    ):
        """
        This yields a score for the data using methods from the spreg library, given the labels provided. If no labels are provided,
        and the object has been fit, then the labels discovered from the previous fit are used.

        If a quorum is not passed, it is assumed to be irrelevant.

        If a quorum is passed and the labels do not meet quorum, the score is inf.

        data        :   (N,P) array of data on which to compute the score of the regions expressed in labels
        data_reg    :   list containing:
                        1- a callable spreg regression method (ex. OLS or GM_Lag)
                        2- np.ndarray of (N,1) shape with N observations on the depedent variable for the regression
                        3- np.ndarray of (N,k) shape with N observations and k columns containing the explanatory variables (constant must not be included)
                        4- pysal W object to be used in the regression (optional)
        all_labels  :   (N,) flat vector of labels expressing the classification of each observation into a region considering the cut under evaluation.
        quorum      :   int expressing the minimum size of regions. Can be -inf if there is no lower bound.
                        Any region below quorum makes the score inf.
        current_labels: (N,) flat vector of labels expressing the classification of each observation into a region not considering the cut under evaluation.

        current_tree: integer indicating the tree label is currently being considered for division
        """

        labels, subtree_quorums = self._prep_score(
            all_labels, current_tree, current_labels
        )
        if (subtree_quorums < quorum).any():
            return np.inf, None
        set_labels = set(labels)
        if data_reg is not None:
            kargs = {
                k: v
                for k, v in data_reg.items()
                if k not in ["reg", "y", "x", "w", "x_nd"]
            }
            trees_scores = {}
            l_arrays = np.array(all_labels)
            w_regi_i = None
            for l in set_labels:
                x = data_reg["x"][all_labels == l]
                if np.linalg.matrix_rank(x) < x.shape[1]:
                    x = np.delete(
                        x, np.nonzero(np.abs(np.diag(np.linalg.qr(x)[1])) < 1e-10), 1
                    )

                if "w" not in data_reg:
                    try:
                        x = np.hstack((np.ones((x.shape[0], 1)), x))
                        reg = data_reg["reg"](
                            y=data_reg["y"][all_labels == l], x=x, **kargs
                        )
                    except np.linalg.LinAlgError:
                        x = np.delete(x, np.nonzero(np.ptp(x, axis=0) == 0), 1)
                        x = np.hstack((np.ones((x.shape[0], 1)), x))
                        reg = data_reg["reg"](
                            y=data_reg["y"][all_labels == l], x=x, **kargs
                        )

                else:
                    regi_ids = list(np.where(l_arrays == l)[0])
                    w_ids = list(map(data_reg["w"].id_order.__getitem__, regi_ids))
                    w_regi_i = w_subset(data_reg["w"], w_ids, silence_warnings=True)
                    try:
                        x = np.hstack((np.ones((x.shape[0], 1)), x))
                        reg = data_reg["reg"](
                            y=data_reg["y"][all_labels == l], x=x, w=w_regi_i, **kargs
                        )
                    except np.linalg.LinAlgError:
                        x = np.delete(x, np.nonzero(np.ptp(x, axis=0) == 0), 1)
                        reg = data_reg["reg"](
                            y=data_reg["y"][all_labels == l], x=x, w=w_regi_i, **kargs
                        )
                trees_scores[l] = np.sum(reg.u ** 2)
            score = sum(trees_scores.values())
        else:
            part_scores, score, trees_scores = self._data_reg_none(
                data, all_labels, l, set_labels
            )

        return score, trees_scores

    def score_stats(
        self,
        data=None,
        data_reg=None,
        all_labels=None,
        quorum=-np.inf,
        current_labels=None,
        current_tree=None,
    ):
        """
        This yields a score for the data using methods from the stats_models library, given the labels provided. If no labels are provided,
        and the object has been fit, then the labels discovered from the previous fit are used.

        If a quorum is not passed, it is assumed to be irrelevant.

        If a quorum is passed and the labels do not meet quorum, the score is inf.

        data        :   (N,P) array of data on which to compute the score of the regions expressed in labels
        data_reg    :   list containing:
                        1- a callable statsmodels regression method (ex. OLS)
                        2- np.ndarray of (N,1) shape with N observations on the depedent variable for the regression
                        3- np.ndarray of (N,k) shape with N observations and k columns containing the explanatory variables (constant must not be included)
                        4- pysal W object to be used in the regression (optional)
        all_labels  :   (N,) flat vector of labels expressing the classification of each observation into a region considering the cut under evaluation.
        quorum      :   int expressing the minimum size of regions. Can be -inf if there is no lower bound.
                        Any region below quorum makes the score inf.
        current_labels: (N,) flat vector of labels expressing the classification of each observation into a region not considering the cut under evaluation.

        current_tree: integer indicating the tree label is currently being considered for division

        NOTE: Optimization occurs with respect to a *dissimilarity* metric, so the problem *minimizes*
              the map dissimilarity. So, lower scores are better.
        """
        labels, subtree_quorums = self._prep_score(
            all_labels, current_tree, current_labels
        )
        if (subtree_quorums < quorum).any():
            return np.inf, None
        set_labels = set(labels)
        if data_reg is not None:
            kargs = {
                k: v
                for k, v in data_reg.items()
                if k not in ["reg", "y", "x", "w", "x_nd"]
            }
            trees_scores = {}
            l_arrays = np.array(all_labels)
            w_regi_i = None
            for l in set_labels:
                x = data_reg["x"][all_labels == l]
                if np.linalg.matrix_rank(x) < x.shape[1]:
                    x = np.delete(
                        x, np.nonzero(np.abs(np.diag(np.linalg.qr(x)[1])) < 1e-10), 1
                    )

                try:
                    x = np.hstack((np.ones((x.shape[0], 1)), x))
                    reg = data_reg["reg"](
                        data_reg["y"][all_labels == l], x, **kargs
                    ).fit()
                except np.linalg.LinAlgError:
                    x = np.delete(x, np.nonzero(np.ptp(x, axis=0) == 0), 1)
                    x = np.hstack((np.ones((x.shape[0], 1)), x))
                    reg = data_reg["reg"](
                        data_reg["y"][all_labels == l], x, **kargs
                    ).fit()

                trees_scores[l] = np.sum(reg.resid ** 2)
            score = sum(trees_scores.values())
        else:
            part_scores, score, trees_scores = self._data_reg_none(
                data, all_labels, l, set_labels
            )
        return score, trees_scores

    def _prep_score(self, all_labels, current_tree, current_labels):
        if all_labels is None:
            try:
                labels = self.current_labels_
            except AttributeError:
                raise ValueError(
                    "Labels not provided and MSF_Prune object has not been fit to data yet."
                )
        if current_tree is not None:
            labels = all_labels[current_labels == current_tree]
        _, subtree_quorums = np.unique(labels, return_counts=True)
        return labels, subtree_quorums

    def _data_reg_none(self, data, all_labels, l, set_labels):
        assert data.shape[0] == len(
            all_labels
        ), "Length of label array ({}) does not match " "length of data ({})! ".format(
            all_labels.shape[0], data.shape[0]
        )
        part_scores = [
            self.reduction(
                self.metric(
                    X=data[all_labels == l],
                    Y=self.center(data[all_labels == l], axis=0).reshape(1, -1),
                )
            )
            for l in set_labels
        ]

        score = self.reduction(part_scores).item()
        trees_scores = {l: part_scores[i] for i, l in enumerate(set_labels)}
        return part_scores, score, trees_scores

    def find_cut(
        self,
        MSF,
        data=None,
        data_reg=None,
        current_n_subtrees=None,
        current_labels=None,
        quorum=-np.inf,
        trees_scores=None,
        labels=None,
        target_label=None,
        make=False,
        verbose=False,
        model_family="spreg",
    ):
        """
        Find the best cut from the MSF.

        MSF: (N,N) scipy sparse matrix with zero elements removed.
             Represents the adjacency matrix for the minimum spanning forest.
             Constructed from sparse.csgraph.sparse_from_dense or using MSF.eliminate_zeros().
             You MUST remove zero entries for this to work, otherwise they are considered no-cost paths.
        data: (N,p) attribute matrix. If not provided, replaced with (N,1) vector of ones.
        data_reg: optional list containing:
                        1- a callable spreg or statsmodels regression method (ex. OLS or GM_Lag)
                        2- np.ndarray of (N,1) shape with N observations on the depedent variable for the regression
                        3- np.ndarray of (N,k) shape with N observations and k columns containing the explanatory variables (constant must not be included)
                        4- pysal W object to be used in the regression (optional)
        current_n_subtrees: integer indication the current number of subtrees.
        current_labels: (N,) flat vector of labels expressing the classification of each observation into a region not considering the cut under evaluation.
        quorum: int denoting the minimum number of elements in the region
        trees_scores: dictionary indicating subtress's labels and their respective current score.
        labels: (N,) flat vector of labels for each point. Represents the "cluster labels"
                for disconnected components of the graph.
        target_label: int from the labels array to subset the MSF. If passed along with `labels`, then a cut
                      will be found that is restricted to that subset of the MSF.
        make: bool, whether or not to modify the input MSF in order to make the best cut that was found.
        verbose: bool/int, denoting how much output to provide to the user, in terms
                 of print statements or progressbars

        Returns a namedtuple with in_node, out_node, and score.
        """
        if data is None:
            data = np.ones(MSF.shape)

        if (labels is None) != (target_label is None):
            raise ValueError(
                "Both labels and target_label must be supplied! Only {} provided.".format(
                    ["labels", "target_label"][int(target_label is None)]
                )
            )
        if verbose:
            try:
                from tqdm import tqdm
            except ImportError:

                def tqdm(noop, desc=""):
                    return noop

        else:

            def tqdm(noop, desc=""):
                return noop

        zero_in = (labels is not None) and (target_label is not None)
        best_deletion = deletion(np.nan, np.nan, np.inf)
        best_d_score = -np.inf
        try:
            old_score = sum(trees_scores.values())
        except:
            pass
        best_scores = {}
        current_list = current_labels.tolist()
        for in_node, out_node in tqdm(
            np.vstack(MSF.nonzero()).T, desc="finding cut..."
        ):  # iterate over MSF edges
            if zero_in:
                if labels[in_node] != target_label:
                    continue

            local_MSF = copy.deepcopy(MSF)
            # delete a candidate edge
            local_MSF[in_node, out_node] = 0
            local_MSF.eliminate_zeros()
            current_tree = current_labels[in_node]

            # get the connected components
            local_n_subtrees, local_labels = cg.connected_components(
                local_MSF, directed=False
            )

            if local_n_subtrees <= current_n_subtrees:
                raise Exception("Malformed MSF!")

            # compute the score of these components
            if model_family == "spreg":
                new_score, new_trees_scores = self.score_spreg(
                    data, data_reg, local_labels, quorum, current_labels, current_tree
                )
            elif model_family == "statsmodels":
                new_score, new_trees_scores = self.score_stats(
                    data, data_reg, local_labels, quorum, current_labels, current_tree
                )
            else:
                raise ValueError("Model family must be either spreg or statsmodels.")

            if np.isfinite(new_score):
                try:
                    d_score = trees_scores[current_tree] - new_score
                    score = old_score - d_score
                except:
                    d_score = -new_score
                    score = new_score
                # if the d_score is greater than the best score and quorum is met
                if d_score > best_d_score:
                    best_deletion = deletion(in_node, out_node, score)
                    best_d_score = d_score
                    try:
                        for i in set(current_labels):
                            best_scores[
                                local_labels[current_list.index(i)]
                            ] = trees_scores[i]
                        for i in new_trees_scores:
                            best_scores[i] = new_trees_scores[i]
                    except:
                        best_scores = new_trees_scores
                    best_MSF = local_MSF
                    best_labels = local_labels
        try:
            return best_deletion, best_scores, best_MSF, local_n_subtrees, best_labels
        except UnboundLocalError:  # in case no solution is found
            return deletion(None, None, np.inf), np.inf, None, np.inf, None


def _const_x(x):
    x = np.delete(x, np.nonzero(np.ptp(x, axis=0) == 0), 1)
    x = np.hstack((np.ones((x.shape[0], 1)), x))
    return x



"""Internal helper files for user output."""

__author__ = "Luc Anselin, Pedro V. Amaral"

import textwrap as TW
import numpy as np
import pandas as pd
from . import diagnostics as diagnostics
from . import diagnostics_tsls as diagnostics_tsls
from . import diagnostics_sp as diagnostics_sp

__all__ = []

###############################################################################
############### Primary functions for running summary diagnostics #############
###############################################################################

def output(reg, vm, other_end=False, robust=False, latex=False):
    if latex:
        print("Warning: Latex output not implemented yet. Using standard output instead.")
    strSummary = output_start(reg)
    for eq in reg.output['equation'].unique():
        try:
            reg.multi[eq].__summary = {}
            strSummary, reg.multi[eq] = out_part_top(strSummary, reg, eq)
        except:
            eq = None
            strSummary, reg = out_part_top(strSummary, reg, eq)
        strSummary, reg = out_part_middle(strSummary, reg, robust, m=eq)
    strSummary, reg = out_part_end(strSummary, reg, vm, other_end, m=eq)
    reg.summary = strSummary
    reg.output.sort_values(by=['equation', 'regime'], inplace=True)
    reg.output.drop(['var_type', 'regime', 'equation'], axis=1, inplace=True)

def output_start(reg):
    reg.__summary = {}
    strSummary = "REGRESSION RESULTS\n"
    strSummary += "------------------\n"
    reg.output = reg.output.assign(coefficients=[None] * len(reg.output), std_err=[None] * len(reg.output),
                                   zt_stat=[None] * len(reg.output), prob=[None] * len(reg.output))
    return strSummary

def out_part_top(strSummary, reg, m):
    # Top part of summary output.
    # m = None for single models, m = 1,2,3... for multiple equation models
    if m == None:
        _reg = reg  # _reg = local object with regression results
    else:
        _reg = reg.multi[m]  # _reg = local object with equation specific regression results
    title = "\nSUMMARY OF OUTPUT: " + _reg.title + "\n"
    strSummary += title
    strSummary += "-" * (len(title) - 2) + "\n"
    strSummary += "%-20s:%12s\n" % ("Data set", _reg.name_ds)
    try:
        strSummary += "%-20s:%12s\n" % ("Weights matrix", _reg.name_w)
    except:
        pass

    strSummary += "%-20s:%12s                %-22s:%12d\n" % (
        "Dependent Variable",
        _reg.name_y,
        "Number of Observations",
        _reg.n,
    )

    strSummary += "%-20s:%12.4f                %-22s:%12d\n" % (
        "Mean dependent var",
        _reg.mean_y,
        "Number of Variables",
        _reg.k,
    )
    strSummary += "%-20s:%12.4f                %-22s:%12d\n" % (
        "S.D. dependent var",
        _reg.std_y,
        "Degrees of Freedom",
        _reg.n - _reg.k,
    )

    _reg.std_err = diagnostics.se_betas(_reg)
    if 'OLS' in reg.__class__.__name__:
        _reg.t_stat = diagnostics.t_stat(_reg)
        _reg.r2 = diagnostics.r2(_reg)
        _reg.ar2 = diagnostics.ar2(_reg)
        strSummary += "%-20s:%12.4f\n%-20s:%12.4f\n" % (
            "R-squared",
            _reg.r2,
            "Adjusted R-squared",
            _reg.ar2,
        )
        _reg.__summary["summary_zt"] = "t"

    else:
        _reg.z_stat = diagnostics.t_stat(_reg, z_stat=True)
        _reg.pr2 = diagnostics_tsls.pr2_aspatial(_reg)
        strSummary += "%-20s:%12.4f\n" % ("Pseudo R-squared", _reg.pr2)
        _reg.__summary["summary_zt"] = "z"

    try:  # Adding additional top part if there is one
        strSummary += _reg.other_top
    except:
        pass

    return (strSummary, _reg)

def out_part_middle(strSummary, reg, robust, m=None):
    # Middle part of summary output.
    # m = None for single models, m = 1,2,3... for multiple equation models
    if m==None:
        _reg = reg #_reg = local object with regression results
        m = reg.output['equation'].unique()[0]
    else:
        _reg = reg.multi[m] #_reg = local object with equation specific regression results
    coefs = pd.DataFrame(_reg.betas, columns=['coefficients'])
    coefs['std_err'] = pd.DataFrame(_reg.std_err)
    try:
        coefs = pd.concat([coefs, pd.DataFrame(_reg.z_stat, columns=['zt_stat', 'prob'])], axis=1)
    except AttributeError:
        coefs = pd.concat([coefs, pd.DataFrame(_reg.t_stat, columns=['zt_stat', 'prob'])], axis=1)
    coefs.index = reg.output[reg.output['equation'] == m].index
    reg.output.update(coefs)
    strSummary += "\n"
    if robust:
        if robust == "white":
            strSummary += "White Standard Errors\n"
        elif robust == "hac":
            strSummary += "HAC Standard Errors; Kernel Weights: " + _reg.name_gwk + "\n"
        elif robust == "ogmm":
            strSummary += "Optimal GMM used to estimate the coefficients and the variance-covariance matrix\n"            
    strSummary += "------------------------------------------------------------------------------------\n"
    strSummary += (
            "            Variable     Coefficient       Std.Error     %1s-Statistic     Probability\n"
            % (_reg.__summary["summary_zt"])
    )
    strSummary += "------------------------------------------------------------------------------------\n"
    m_output = reg.output[reg.output['equation'] == m]
    for row in m_output.iloc[np.lexsort((m_output.index, m_output['regime']))].itertuples():
        try:
            strSummary += "%20s    %12.5f    %12.5f    %12.5f    %12.5f\n" % (
                row.var_names,
                row.coefficients,
                row.std_err,
                row.zt_stat,
                row.prob
            )
        except TypeError:  # special case for models that do not have inference on the lambda term
            strSummary += "%20s    %12.5f    \n" % (
                row.var_names,
                row.coefficients
            )
    strSummary += "------------------------------------------------------------------------------------\n"

    try:  # Adding info on instruments if they are present
        name_q = _reg.name_q
        name_yend = _reg.name_yend
        insts = "Instruments: "
        for name in sorted(name_q):
            insts += name + ", "
        text_wrapper = TW.TextWrapper(width=76, subsequent_indent="             ")
        insts = text_wrapper.fill(insts[:-2])
        insts += "\n"
        inst2 = "Instrumented: "
        for name in sorted(name_yend):
            inst2 += name + ", "
        text_wrapper = TW.TextWrapper(width=76, subsequent_indent="             ")
        inst2 = text_wrapper.fill(inst2[:-2])
        inst2 += "\n"
        inst2 += insts
        strSummary += inst2
    except:
        pass

    try:  # Adding info on regimes if they are present
        strSummary += ("Regimes variable: %s\n" % _reg.name_regimes)
        strSummary += _summary_chow(_reg)  # If local regimes present, compute Chow test.
    except:
        pass

    try:  # Adding local warning if there is one
        if _reg.warning:
            strSummary += _reg.warning
    except:
        pass

    try:  # Adding other middle part if there are any
        strSummary += _reg.other_mid
    except:
        pass

    return (strSummary, reg)

def out_part_end(strSummary, reg, vm, other_end, m=None):
    if m is not None:
        strSummary += "------------------------------------------------------------------------------------\n"
        try:  # Adding global warning if there is one
            strSummary += reg.warning
        except:
            pass
        strSummary += "\nGLOBAL DIAGNOSTICS\n"
        try:  # Adding global Chow test if there is one
            strSummary += _summary_chow(reg)
        except:
            pass
    if other_end:
        strSummary += other_end
    if vm:
        strSummary += _summary_vm(reg)
    strSummary += "================================ END OF REPORT ====================================="
    return (strSummary, reg)

def _summary_chow(reg):
    sum_text = "\nREGIMES DIAGNOSTICS - CHOW TEST"
    name_x_r = reg.name_x_r
    joint, regi = reg.chow.joint, reg.chow.regi
    sum_text += "\n                 VARIABLE        DF        VALUE           PROB\n"
    if reg.cols2regi == "all":
        names_chow = name_x_r[1:]
    else:
        names_chow = [name_x_r[1:][i] for i in np.where(reg.cols2regi)[0]]
    if reg.constant_regi == "many":
        names_chow = ["CONSTANT"] + names_chow

    if 'lambda' in reg.output.var_type.values:
        if reg.output.var_type.value_counts()['lambda'] > 1:
            names_chow += ["lambda"]

    reg.output_chow = pd.DataFrame()
    reg.output_chow['var_names'] = names_chow
    reg.output_chow['df'] = reg.nr - 1
    reg.output_chow = pd.concat([reg.output_chow, pd.DataFrame(regi, columns=['value', 'prob'])], axis=1)
    reg.output_chow = pd.concat([reg.output_chow, pd.DataFrame([{'var_names': 'Global test',
                                              'df':  reg.kr * (reg.nr - 1),
                                              'value': joint[0], 'prob': joint[1]}])], ignore_index=True)
    for row in reg.output_chow.itertuples():
        sum_text += "%25s        %2d    %12.3f        %9.4f\n" % (
            row.var_names,
            row.df,
            row.value,
            row.prob
        )

    return sum_text


def _spat_diag_out(reg, w, type, moran=False):
    strSummary = "\nDIAGNOSTICS FOR SPATIAL DEPENDENCE\n"
    if not moran:
        strSummary += (
            "TEST                              DF       VALUE           PROB\n"
        )
    else:
        strSummary += (
            "TEST                           MI/DF       VALUE           PROB\n"
        )

    cache = diagnostics_sp.spDcache(reg, w)
    if type == "yend":
        mi, ak, ak_p = diagnostics_sp.akTest(reg, w, cache)
        reg.ak_test = ak, ak_p
        strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
            "Anselin-Kelejian Test",
            1,
            reg.ak_test[0],
            reg.ak_test[1],
        )
    elif type == "ols":
        lm_tests = diagnostics_sp.LMtests(reg, w)
        reg.lm_error = lm_tests.lme
        reg.lm_lag = lm_tests.lml
        reg.rlm_error = lm_tests.rlme
        reg.rlm_lag = lm_tests.rlml
        reg.lm_sarma = lm_tests.sarma
        if moran:
            moran_res = diagnostics_sp.MoranRes(reg, w, z=True)
            reg.moran_res = moran_res.I, moran_res.zI, moran_res.p_norm
            strSummary += "%-27s  %8.4f     %9.3f        %9.4f\n" % (
                "Moran's I (error)",
                reg.moran_res[0],
                reg.moran_res[1],
                reg.moran_res[2],
            )
        strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
            "Lagrange Multiplier (lag)",
            1,
            reg.lm_lag[0],
            reg.lm_lag[1],
        )
        strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
            "Robust LM (lag)",
            1,
            reg.rlm_lag[0],
            reg.rlm_lag[1],
        )
        strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
            "Lagrange Multiplier (error)",
            1,
            reg.lm_error[0],
            reg.lm_error[1],
        )
        strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
            "Robust LM (error)",
            1,
            reg.rlm_error[0],
            reg.rlm_error[1],
        )
        strSummary += "%-27s      %2d    %12.3f        %9.4f\n\n" % (
            "Lagrange Multiplier (SARMA)",
            2,
            reg.lm_sarma[0],
            reg.lm_sarma[1],
        )
    return strSummary

def _nonspat_top(reg, ml=False):
    if not ml:
        reg.sig2ML = reg.sig2n
        reg.f_stat = diagnostics.f_stat(reg)
        reg.logll = diagnostics.log_likelihood(reg)
        reg.aic = diagnostics.akaike(reg)
        reg.schwarz = diagnostics.schwarz(reg)

        strSummary = "%-20s:%12.6g                %-22s:%12.4f\n" % (
            "Sum squared residual", reg.utu, "F-statistic", reg.f_stat[0],)
        strSummary += "%-20s:%12.3f                %-22s:%12.4g\n" % (
            "Sigma-square",
            reg.sig2,
            "Prob(F-statistic)",
            reg.f_stat[1],
        )
        strSummary += "%-20s:%12.3f                %-22s:%12.3f\n" % (
            "S.E. of regression",
            np.sqrt(reg.sig2),
            "Log likelihood",
            reg.logll,
        )
        strSummary += "%-20s:%12.3f                %-22s:%12.3f\n" % (
            "Sigma-square ML",
            reg.sig2ML,
            "Akaike info criterion",
            reg.aic,
        )
        strSummary += "%-20s:%12.4f                %-22s:%12.3f\n" % (
            "S.E of regression ML",
            np.sqrt(reg.sig2ML),
            "Schwarz criterion",
            reg.schwarz,
        )
    else:
        strSummary = "%-20s:%12.4f\n" % (
            "Log likelihood",
            reg.logll,
        )
        strSummary += "%-20s:%12.4f                %-22s:%12.3f\n" % (
            "Sigma-square ML",
            reg.sig2,
            "Akaike info criterion",
            reg.aic,
        )
        strSummary += "%-20s:%12.4f                %-22s:%12.3f\n" % (
            "S.E of regression",
            np.sqrt(reg.sig2),
            "Schwarz criterion",
            reg.schwarz,
        )

    return strSummary

def _nonspat_mid(reg, white_test=False):
    # compute diagnostics
    reg.mulColli = diagnostics.condition_index(reg)
    reg.jarque_bera = diagnostics.jarque_bera(reg)
    reg.breusch_pagan = diagnostics.breusch_pagan(reg)
    reg.koenker_bassett = diagnostics.koenker_bassett(reg)

    if white_test:
        reg.white = diagnostics.white(reg)

    strSummary = "\nREGRESSION DIAGNOSTICS\n"
    if reg.mulColli:
        strSummary += "MULTICOLLINEARITY CONDITION NUMBER %16.3f\n\n" % (reg.mulColli)
    strSummary += "TEST ON NORMALITY OF ERRORS\n"
    strSummary += "TEST                             DF        VALUE           PROB\n"
    strSummary += "%-27s      %2d  %14.3f        %9.4f\n\n" % (
        "Jarque-Bera",
        reg.jarque_bera["df"],
        reg.jarque_bera["jb"],
        reg.jarque_bera["pvalue"],
    )
    strSummary += "DIAGNOSTICS FOR HETEROSKEDASTICITY\n"
    strSummary += "RANDOM COEFFICIENTS\n"
    strSummary += "TEST                             DF        VALUE           PROB\n"
    strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
        "Breusch-Pagan test",
        reg.breusch_pagan["df"],
        reg.breusch_pagan["bp"],
        reg.breusch_pagan["pvalue"],
    )
    strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
        "Koenker-Bassett test",
        reg.koenker_bassett["df"],
        reg.koenker_bassett["kb"],
        reg.koenker_bassett["pvalue"],
    )
    try:
        if reg.white:
            strSummary += "\nSPECIFICATION ROBUST TEST\n"
            if len(reg.white) > 3:
                strSummary += reg.white + "\n"
            else:
                strSummary += (
                    "TEST                             DF        VALUE           PROB\n"
                )
                strSummary += "%-27s      %2d    %12.3f        %9.4f\n" % (
                    "White",
                    reg.white["df"],
                    reg.white["wh"],
                    reg.white["pvalue"],
                )
    except:
        pass
    return strSummary

def _spat_pseudo_r2(reg):
    if np.abs(reg.rho) < 1:
        reg.pr2_e = diagnostics_tsls.pr2_spatial(reg)
        strSummary = "%-20s:  %5.4f\n" % ("Spatial Pseudo R-squared", reg.pr2_e)
    else:
        strSummary =  "Spatial Pseudo R-squared: omitted due to rho outside the boundary (-1, 1).\n"
    return strSummary

def _summary_vm(reg):
    strVM = "\n"
    strVM += "COEFFICIENTS VARIANCE MATRIX\n"
    strVM += "----------------------------\n"
    try:
        for name in reg.name_z:
            strVM += "%12s" % (name)
    except:
        for name in reg.name_x:
            strVM += "%12s" % (name)
    strVM += "\n"
    nrow = reg.vm.shape[0]
    ncol = reg.vm.shape[1]
    for i in range(nrow):
        for j in range(ncol):
            strVM += "%12.6f" % (reg.vm[i][j])
        strVM += "\n"
    return strVM

def _summary_iteration(reg):
    """Reports the number of iterations computed and the type of estimator used for hom and het models."""
    try:
        niter = reg.niter
    except:
        niter = reg.iteration

    txt = "%-20s:%12s\n" % ("N. of iterations", niter)

    try:
        if reg.step1c:
            step1c = "Yes"
        else:
            step1c = "No"
        txt = txt[:-1] + "                %-22s:%12s\n" % (
            "Step1c computed",
            step1c,
        )
    except:
       pass

    try:
        txt = txt[:-1] + "                %-22s:%12s" % (
            "A1 type: ",
            reg.A1,
        )
    except:
        pass

    return txt



import unittest
import numpy as np
import libpysal
from spreg.twosls_sp_regimes import GM_Lag_Regimes
from spreg import utils
from spreg.twosls_sp import GM_Lag
from libpysal.common import RTOL

class TestGMLag_Regimes(unittest.TestCase):
    def setUp(self):
        self.w = libpysal.weights.Queen.from_shapefile(libpysal.examples.get_path("columbus.shp"))
        self.w.transform = 'r'
        self.db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), 'r')
        y = np.array(self.db.by_col("CRIME"))
        self.y = np.reshape(y, (49,1))
        self.r_var = 'NSA'
        self.regimes = self.db.by_col(self.r_var)

    def test___init__(self):
        #Matches SpaceStat
        X = []
        X.append(self.db.by_col("INC"))
        X.append(self.db.by_col("HOVAL"))
        self.X = np.array(X).T
        reg = GM_Lag_Regimes(self.y, self.X, self.regimes, w=self.w, sig2n_k=True, regime_lag_sep=False, regime_err_sep=False) 
        betas = np.array([[ 45.14892906],
       [ -1.42593383],
       [ -0.11501037],
       [ 40.99023016],
       [ -0.81498302],
       [ -0.28391409],
       [  0.4736163 ]])
        np.testing.assert_allclose(reg.betas, betas,RTOL)
        e_5 = np.array([[ -1.47960519],
       [ -7.93748769],
       [ -5.88561835],
       [-13.37941105],
       [  5.2524303 ]])
        np.testing.assert_allclose(reg.e_pred[0:5], e_5,RTOL)
        h_0 = np.array([[  0.       ,   0.       ,   0.       ,   1.       ,  19.531    ,
         80.467003 ,   0.       ,   0.       ,  18.594    ,  35.4585005]])
        np.testing.assert_allclose(reg.h[0]*np.eye(10), h_0)
        self.assertEqual(reg.k, 7)
        self.assertEqual(reg.kstar, 1)
        np.testing.assert_allclose(reg.mean_y, 35.128823897959187,RTOL)
        self.assertEqual(reg.n, 49)
        np.testing.assert_allclose(reg.pr2, 0.6572182131915739,RTOL)
        np.testing.assert_allclose(reg.pr2_e, 0.5779687278635434,RTOL)
        pfora1a2 = np.array([ -2.15017629,  -0.30169328,  -0.07603704, -22.06541809,
         0.45738058,   0.02805828,   0.39073923]) 
        np.testing.assert_allclose(reg.pfora1a2[0], pfora1a2,RTOL)
        predy_5 = np.array([[ 13.93216104],
       [ 23.46424269],
       [ 34.43510955],
       [ 44.32473878],
       [ 44.39117516]])
        np.testing.assert_allclose(reg.predy[0:5], predy_5,RTOL)
        predy_e_5 = np.array([[ 17.20558519],
       [ 26.73924169],
       [ 36.51239935],
       [ 45.76717105],
       [ 45.4790797 ]])
        np.testing.assert_allclose(reg.predy_e[0:5], predy_e_5,RTOL)
        q_5 = np.array([[  0.       ,   0.       ,  18.594    ,  35.4585005]])
        np.testing.assert_allclose(reg.q[0]*np.eye(4), q_5, RTOL)
        self.assertEqual(reg.robust, 'unadjusted')
        np.testing.assert_allclose(reg.sig2n_k, 109.76462904625834,RTOL)
        np.testing.assert_allclose(reg.sig2n, 94.08396775393571,RTOL)
        np.testing.assert_allclose(reg.sig2, 109.76462904625834,RTOL)
        np.testing.assert_allclose(reg.std_y, 16.732092091229699,RTOL)
        u_5 = np.array([[  1.79381896],
       [ -4.66248869],
       [ -3.80832855],
       [-11.93697878],
       [  6.34033484]])
        np.testing.assert_allclose(reg.u[0:5], u_5,RTOL)
        np.testing.assert_allclose(reg.utu, 4610.11441994285,RTOL)
        varb = np.array([  1.23841820e+00,  -3.65620114e-02,  -1.21919663e-03,
         1.00057547e+00,  -2.07403182e-02,  -1.27232693e-03,
        -1.77184084e-02])
        np.testing.assert_allclose(reg.varb[0], varb,RTOL)
        vm = np.array([  1.35934514e+02,  -4.01321561e+00,  -1.33824666e-01,
         1.09827796e+02,  -2.27655334e+00,  -1.39656494e-01,
        -1.94485452e+00])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        x_0 = np.array([[  0.      ,   0.      ,   0.      ,   1.      ,  19.531   ,
         80.467003]])
        np.testing.assert_allclose(reg.x[0]*np.eye(6), x_0,RTOL)
        y_5 = np.array([[ 15.72598 ],
       [ 18.801754],
       [ 30.626781],
       [ 32.38776 ],
       [ 50.73151 ]])
        np.testing.assert_allclose(reg.y[0:5], y_5,RTOL)
        yend_5 = np.array([[ 24.7142675 ],
       [ 26.24684033],
       [ 29.411751  ],
       [ 34.64647575],
       [ 40.4653275 ]])
        np.testing.assert_allclose(reg.yend[0:5]*np.array([[1]]), yend_5,RTOL)
        z_0 = np.array([[  0.       ,   0.       ,   0.       ,   1.       ,  19.531    ,
         80.467003 ,  24.7142675]]) 
        np.testing.assert_allclose(reg.z[0]*np.eye(7), z_0,RTOL)
        zthhthi = np.array([  1.00000000e+00,  -2.35922393e-16,   5.55111512e-17,
         0.00000000e+00,   0.00000000e+00,   0.00000000e+00,
        -4.44089210e-16,   2.22044605e-16,   0.00000000e+00,
         0.00000000e+00])
        # np.testing.assert_allclose(reg.zthhthi[0], zthhthi, RTOL)
        np.testing.assert_array_almost_equal(reg.zthhthi[0], zthhthi)
        chow_regi = np.array([[ 0.19692667,  0.65721307],
       [ 0.5666492 ,  0.45159351],
       [ 0.45282066,  0.5009985 ]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,RTOL)
        np.testing.assert_allclose(reg.chow.joint[0], 0.82409867601863462,RTOL)
    
    def test_init_discbd(self):
        #Matches SpaceStat.
        X = np.array(self.db.by_col("INC"))
        X = np.reshape(X, (49,1))
        yd = np.array(self.db.by_col("HOVAL"))
        yd = np.reshape(yd, (49,1))
        q = np.array(self.db.by_col("DISCBD"))
        q = np.reshape(q, (49,1))
        reg = GM_Lag_Regimes(self.y, X, self.regimes, yend=yd, q=q, lag_q=False, w=self.w, sig2n_k=True, regime_lag_sep=False, regime_err_sep=False) 
        tbetas = np.array([[ 42.7266306 ],
       [ -0.15552345],
       [ 37.70545276],
       [ -0.5341577 ],
       [ -0.68305796],
       [ -0.37106077],
       [  0.55809516]])
        np.testing.assert_allclose(tbetas, reg.betas,RTOL)
        vm = np.array([ 270.62979422,    3.62539081,  327.89638627,    6.24949355,
         -5.25333106,   -6.01743515,   -4.19290074])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        e_3 = np.array([[-0.33142796],
       [-9.51719607],
       [-7.86272153]])
        np.testing.assert_allclose(reg.e_pred[0:3], e_3,RTOL)
        u_3 = np.array([[ 4.51839601],
       [-5.67363147],
       [-5.1927562 ]])
        np.testing.assert_allclose(reg.u[0:3], u_3,RTOL)
        predy_3 = np.array([[ 11.20758399],
       [ 24.47538547],
       [ 35.8195372 ]])
        np.testing.assert_allclose(reg.predy[0:3], predy_3,RTOL)
        predy_e_3 = np.array([[ 16.05740796],
       [ 28.31895007],
       [ 38.48950253]])
        np.testing.assert_allclose(reg.predy_e[0:3], predy_e_3,RTOL)
        chow_regi = np.array([[ 0.13130991,  0.71707772],
       [ 0.04740966,  0.82763357],
       [ 0.15474413,  0.6940423 ]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,RTOL)
        np.testing.assert_allclose(reg.chow.joint[0], 0.31248100032096549,RTOL)
    
    def test_lag_q(self):
        X = np.array(self.db.by_col("INC"))
        X = np.reshape(X, (49,1))
        yd = np.array(self.db.by_col("HOVAL"))
        yd = np.reshape(yd, (49,1))
        q = np.array(self.db.by_col("DISCBD"))
        q = np.reshape(q, (49,1))
        reg = GM_Lag_Regimes(self.y, X, self.regimes, yend=yd, q=q, w=self.w, sig2n_k=True, regime_lag_sep=False, regime_err_sep=False) 
        tbetas = np.array([[ 37.87698329],
       [ -0.89426982],
       [ 31.4714777 ],
       [ -0.71640525],
       [ -0.28494432],
       [ -0.2294271 ],
       [  0.62996544]])
        np.testing.assert_allclose(tbetas, reg.betas,RTOL)
        vm = np.array([ 128.25714554,   -0.38975354,   95.7271044 ,   -1.8429218 ,
         -1.75331978,   -0.18240338,   -1.67767464])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        chow_regi = np.array([[ 0.43494049,  0.50957463],
       [ 0.02089281,  0.88507135],
       [ 0.01180501,  0.91347943]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,RTOL)
        np.testing.assert_allclose(reg.chow.joint[0], 0.54288190938307757,RTOL)
    
    def test_all_regi(self):
        X = np.array(self.db.by_col("INC"))
        X = np.reshape(X, (49,1))
        yd = np.array(self.db.by_col("HOVAL"))
        yd = np.reshape(yd, (49,1))
        q = np.array(self.db.by_col("DISCBD"))
        q = np.reshape(q, (49,1))
        reg = GM_Lag_Regimes(self.y, X, self.regimes, yend=yd, q=q, w=self.w, regime_lag_sep=False, regime_err_sep=False) 
        tbetas = np.array([[ 37.87698329,  -0.89426982,  31.4714777 ,  -0.71640525,
         -0.28494432,  -0.2294271 ,   0.62996544]])
        np.testing.assert_allclose(tbetas, reg.betas.T,RTOL)
        vm = np.array([109.934696,  -0.334074,  82.051804,  -1.579647,  -1.502846,
        -0.156346,  -1.438007])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        e_3 = np.array([[-2.66997799],
       [-7.69786264],
       [-4.39412782]])
        np.testing.assert_allclose(reg.e_pred[0:3], e_3,RTOL)
        u_3 = np.array([[ 1.13879007],
       [-3.76873198],
       [-1.89671717]])
        np.testing.assert_allclose(reg.u[0:3], u_3,RTOL)
        predy_3 = np.array([[ 14.58718993],
       [ 22.57048598],
       [ 32.52349817]])
        np.testing.assert_allclose(reg.predy[0:3], predy_3,RTOL)
        predy_e_3 = np.array([[ 18.39595799],
       [ 26.49961664],
       [ 35.02090882]])
        np.testing.assert_allclose(reg.predy_e[0:3], predy_e_3,RTOL)
        chow_regi = np.array([[0.507431, 0.476253],
       [0.024375, 0.875935],
       [0.013773, 0.906578]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,1e-04)
        np.testing.assert_allclose(reg.chow.joint[0], 0.633362,RTOL)
    
    def test_all_regi_sig2(self):
        #Artficial:
        n = 256
        x1 = np.random.uniform(-10,10,(n,1))
        x2 = np.random.uniform(1,5,(n,1))
        q = x2 + np.random.normal(0,1,(n,1))
        x = np.hstack((x1,x2))
        y = np.dot(np.hstack((np.ones((n,1)),x)),np.array([[1],[0.5],[2]])) + np.random.normal(0,1,(n,1))
        latt = int(np.sqrt(n))
        w = libpysal.weights.util.lat2W(latt,latt)
        w.transform='r'
        regi = [0]*(n//2) + [1]*(n//2)
        model = GM_Lag_Regimes(y, x1, regi, q=q, yend=x2, w=w, regime_lag_sep=True, regime_err_sep=True)
        w1 = libpysal.weights.util.lat2W(latt//2,latt)
        w1.transform='r'
        model1 = GM_Lag(y[0:(n//2)].reshape((n//2),1), x1[0:(n//2)],yend=x2[0:(n//2)], q=q[0:(n//2)], w=w1)
        model2 = GM_Lag(y[(n//2):n].reshape((n//2),1), x1[(n//2):n],yend=x2[(n//2):n], q=q[(n//2):n], w=w1)
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas,tbetas)
        vm = np.hstack((model1.vm.diagonal(),model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm,RTOL)
        #Columbus:
        X = np.array(self.db.by_col("INC"))
        X = np.reshape(X, (49,1))
        yd = np.array(self.db.by_col("HOVAL"))
        yd = np.reshape(yd, (49,1))
        q = np.array(self.db.by_col("DISCBD"))
        q = np.reshape(q, (49,1))
        reg = GM_Lag_Regimes(self.y, X, self.regimes, yend=yd, q=q, w=self.w,regime_lag_sep=True, regime_err_sep = True) 
        tbetas = np.array([[ 42.35827477],
       [ -0.09472413],
       [ -0.68794223],
       [  0.54482537],
       [ 32.24228762],
       [ -0.12304063],
       [ -0.46840307],
       [  0.67108156]])
        np.testing.assert_allclose(tbetas, reg.betas)
        vm = np.array([ 200.92894859,    4.56244927,   -4.85603079,   -2.9755413 ,
          0.        ,    0.        ,    0.        ,    0.        ])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        e_3 = np.array([[ -1.32209547],
       [-13.15611199],
       [-11.62357696]])
        np.testing.assert_allclose(reg.e_pred[0:3], e_3,RTOL)
        u_3 = np.array([[ 6.99250069],
       [-7.5665856 ],
       [-7.04753328]])
        np.testing.assert_allclose(reg.u[0:3], u_3,RTOL)
        predy_3 = np.array([[  8.73347931],
       [ 26.3683396 ],
       [ 37.67431428]])
        np.testing.assert_allclose(reg.predy[0:3], predy_3,RTOL)
        predy_e_3 = np.array([[ 17.04807547],
       [ 31.95786599],
       [ 42.25035796]])
        np.testing.assert_allclose(reg.predy_e[0:3], predy_e_3,RTOL)
        chow_regi = np.array([[  1.51825373e-01,   6.96797034e-01],
       [  3.20105698e-04,   9.85725412e-01],
       [  8.58836996e-02,   7.69476896e-01],
       [  1.01357290e-01,   7.50206873e-01]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,RTOL)
        np.testing.assert_allclose(reg.chow.joint[0], 0.38417230022512161,RTOL)

    def test_fixed_const(self):
        X = np.array(self.db.by_col("INC"))
        X = np.reshape(X, (49,1))
        yd = np.array(self.db.by_col("HOVAL"))
        yd = np.reshape(yd, (49,1))
        q = np.array(self.db.by_col("DISCBD"))
        q = np.reshape(q, (49,1))
        reg = GM_Lag_Regimes(self.y, X, self.regimes, yend=yd, q=q, w=self.w, constant_regi='one', regime_lag_sep=False, regime_err_sep=False) 
        tbetas = np.array([[ -0.37658823],
       [ -0.9666079 ],
       [ 35.5445944 ],
       [ -0.45793559],
       [ -0.24216904],
       [  0.62500602]])
        np.testing.assert_allclose(tbetas, reg.betas,RTOL)
        vm = np.array([ 1.4183697 , -0.05975784, -0.27161863, -0.62517245,  0.02266177,
        0.00312976])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        e_3 = np.array([[ 0.17317815],
       [-5.53766328],
       [-3.82889307]])
        np.testing.assert_allclose(reg.e_pred[0:3], e_3,RTOL)
        u_3 = np.array([[ 3.10025518],
       [-1.83150689],
       [-1.49598494]])
        np.testing.assert_allclose(reg.u[0:3], u_3,RTOL)
        predy_3 = np.array([[ 12.62572482],
       [ 20.63326089],
       [ 32.12276594]])
        np.testing.assert_allclose(reg.predy[0:3], predy_3,RTOL)
        predy_e_3 = np.array([[ 15.55280185],
       [ 24.33941728],
       [ 34.45567407]])
        np.testing.assert_allclose(reg.predy_e[0:3], predy_e_3,RTOL)
        chow_regi = np.array([[  1.85767047e-01,   6.66463269e-01],
       [  1.19445012e+01,   5.48089036e-04]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,RTOL)
        np.testing.assert_allclose(reg.chow.joint[0], 12.017256217621382,RTOL)

    def test_names(self):
        y_var = 'CRIME'
        x_var = ['INC']
        x = np.array([self.db.by_col(name) for name in x_var]).T
        yd_var = ['HOVAL']
        yd = np.array([self.db.by_col(name) for name in yd_var]).T
        q_var = ['DISCBD']
        q = np.array([self.db.by_col(name) for name in q_var]).T
        r_var = 'NSA'
        reg = GM_Lag_Regimes(self.y, x, self.regimes, yend=yd, q=q, w=self.w, name_y=y_var, name_x=x_var, name_yend=yd_var, name_q=q_var, name_regimes=r_var, name_ds='columbus', name_w='columbus.gal', regime_lag_sep=False, regime_err_sep=False)
        betas = np.array([[ 37.87698329],
       [ -0.89426982],
       [ 31.4714777 ],
       [ -0.71640525],
       [ -0.28494432],
       [ -0.2294271 ],
       [  0.62996544]])
        np.testing.assert_allclose(reg.betas, betas,RTOL)
        vm = np.array([ 109.93469618,   -0.33407447,   82.05180377,   -1.57964725,
         -1.50284553,   -0.15634575,   -1.43800683])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        chow_regi = np.array([[ 0.50743058,  0.47625326],
       [ 0.02437494,  0.87593468],
       [ 0.01377251,  0.9065777 ]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,RTOL)
        np.testing.assert_allclose(reg.chow.joint[0], 0.63336222761359162,RTOL)
        self.assertListEqual(reg.name_x, ['0_CONSTANT', '0_INC', '1_CONSTANT', '1_INC'])
        self.assertListEqual(reg.name_yend, ['0_HOVAL', '1_HOVAL', '_Global_W_CRIME'])
        self.assertListEqual(reg.name_q, ['0_DISCBD', '0_W_INC', '0_W_DISCBD', '1_DISCBD', '1_W_INC', '1_W_DISCBD'])
        self.assertEqual(reg.name_y, y_var)

if __name__ == '__main__':
    unittest.main()



import unittest
import libpysal
import numpy as np
import pandas as pd
from spreg.panel_re import Panel_RE_Lag, Panel_RE_Error
from libpysal.common import RTOL
from libpysal.weights import w_subset


class Test_Panel_RE_Lag(unittest.TestCase):
    def setUp(self):
        self.ds_name = "NCOVR"
        nat = libpysal.examples.load_example(self.ds_name)
        self.db = libpysal.io.open(nat.get_path("NAT.dbf"), "r")
        nat_shp = libpysal.examples.get_path("NAT.shp")
        w_full = libpysal.weights.Queen.from_shapefile(nat_shp)
        self.y_name = ["HR70", "HR80", "HR90"]
        self.x_names = ["RD70", "RD80", "RD90", "PS70", "PS80", "PS90"]
        c_names = ["STATE_NAME", "FIPSNO"]
        y_full = [self.db.by_col(name) for name in self.y_name]
        y_full = np.array(y_full).T
        x_full = [self.db.by_col(name) for name in self.x_names]
        x_full = np.array(x_full).T
        c_full = [self.db.by_col(name) for name in c_names]
        c_full = pd.DataFrame(c_full, index=c_names).T
        filter_states = ["Kansas", "Missouri", "Oklahoma", "Arkansas"]
        filter_counties = c_full[c_full["STATE_NAME"].isin(filter_states)]
        filter_counties = filter_counties["FIPSNO"].values
        counties = np.array(self.db.by_col("FIPSNO"))
        subid = np.where(np.isin(counties, filter_counties))[0]
        self.w = w_subset(w_full, subid)
        self.w.transform = "r"
        self.y = y_full[
            subid,
        ]
        self.x = x_full[
            subid,
        ]

    def test_Panel(self):
        reg = Panel_RE_Lag(
            self.y,
            self.x,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_ds=self.ds_name,
        )
        betas = np.array(
            [[4.44421994], [2.52821717], [2.24768846], [0.25846846], [0.68426639]]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([1.17169293])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([2.43910394])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        vm = np.array([0.08734092, 0.05232857, 0.05814063, 0.00164801, 0.00086908])
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = np.array([[15.71234238]])
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.2634518198611293
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array([0.29553498, 0.22875438, 0.24112368, 0.04059565, 0.02948021])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        logll = -3127.652757262218
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 6263.305514524436
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 6283.3755390962015
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)


class Test_Panel_RE_Error(unittest.TestCase):
    def setUp(self):
        self.ds_name = "NCOVR"
        nat = libpysal.examples.load_example(self.ds_name)
        self.db = libpysal.io.open(nat.get_path("NAT.dbf"), "r")
        nat_shp = libpysal.examples.get_path("NAT.shp")
        w_full = libpysal.weights.Queen.from_shapefile(nat_shp)
        self.y_name = ["HR70", "HR80", "HR90"]
        self.x_names = ["RD70", "RD80", "RD90", "PS70", "PS80", "PS90"]
        c_names = ["STATE_NAME", "FIPSNO"]
        y_full = [self.db.by_col(name) for name in self.y_name]
        y_full = np.array(y_full).T
        x_full = [self.db.by_col(name) for name in self.x_names]
        x_full = np.array(x_full).T
        c_full = [self.db.by_col(name) for name in c_names]
        c_full = pd.DataFrame(c_full, index=c_names).T
        filter_states = ["Kansas", "Missouri", "Oklahoma", "Arkansas"]
        filter_counties = c_full[c_full["STATE_NAME"].isin(filter_states)]
        filter_counties = filter_counties["FIPSNO"].values
        counties = np.array(self.db.by_col("FIPSNO"))
        subid = np.where(np.isin(counties, filter_counties))[0]
        self.w = w_subset(w_full, subid)
        self.w.transform = "r"
        self.y = y_full[
            subid,
        ]
        self.x = x_full[
            subid,
        ]

#    def test_Panel(self):
#        reg = Panel_RE_Error(
#            self.y,
#            self.x,
#            w=self.w,
#            name_y=self.y_name,
#            name_x=self.x_names,
#            name_ds=self.ds_name,
#        )
#        betas = np.array(
#            [[5.87893756], [3.23269025], [2.62996804], [0.34042682], [4.9782446]]
#        )
#        np.testing.assert_allclose(reg.betas, betas, RTOL)
#        u = np.array([-0.2372652])
#        np.testing.assert_allclose(reg.u[0], u, RTOL)
#        predy = np.array([4.27277771])
#        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
#        vm = np.array([0.05163595, 0.05453637, 0.06134783, 0.00025012, 0.0030366])
#        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
#        sig2 = np.array([[16.10231419]])
#        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
#        pr2 = 0.3256008995950422
#        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
#        std_err = np.array([0.22723545, 0.23353024, 0.24768493, 0.01581518, 0.05510535])
#        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
#        logll = -7183.836220934392
#        np.testing.assert_allclose(reg.logll, logll, RTOL)
#        aic = 14373.672441868785
#        np.testing.assert_allclose(reg.aic, aic, RTOL)
#        schwarz = 14388.724960297608
#        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)


if __name__ == "__main__":
    unittest.main()
